import abc
import logging

import numpy as np

from aspire.image import Image
from aspire.image.xform import Xform
from aspire.numeric import fft, xp
from aspire.operators import ArrayFilter, PowerFilter, ScalarFilter
from aspire.utils import grid_2d, randn, trange

logger = logging.getLogger(__name__)


# TODO: Implement correct hierarchy and DRY


class NoiseAdder(Xform):
    """
    Defines interface for `CustomNoiseAdder`s.
    """

    def __init__(self, noise_filter, seed=0):
        """
        Initialize the random state of this `NoiseAdder` using `noise_filter` and `seed`.

        `noise_filter` will be provided by the user or instantiated automatically by the subclass.

        :param seed: The random seed used to generate white noise.
        :param noise_filter: An `aspire.operators.Filter` object.
            `NoiseAdders` start by generating gaussian noise,
            then apply `noise_filter` to transform the noise.
            Note the `noise_filter` will be raised to the 1/2 power.
        """
        super().__init__()
        self.seed = seed
        self._noise_filter = noise_filter
        self.noise_filter = PowerFilter(noise_filter, power=0.5)

    def __repr__(self):
        return f"{self.__class__.__name__}(noise_filter={self._noise_filter}, seed={self.seed})"

    def __str__(self):
        return f"{self.__class__.__name__}"

    def _forward(self, im, indices):
        _im = im.asnumpy().copy()

        for i, idx in enumerate(indices):
            # Note: The following random seed behavior is directly taken from MATLAB Cov3D code.
            random_seed = self.seed + 191 * (idx + 1)
            im_s = randn(2 * im.resolution, 2 * im.resolution, seed=random_seed)
            # Use numpy because im_s and im are different image sizes
            im_s = Image(im_s).filter(self.noise_filter).asnumpy()[0]
            _im[i] += im_s[: im.resolution, : im.resolution]

        return Image(_im)

    @abc.abstractproperty
    def noise_var(self):
        """
        Concrete implementations are expected to provide a method that
        returns the noise variance for the NoiseAdder.

        Authors of `NoiseAdder`s are encouraged to consider any relevant
        methods of calculating noise variance from theory.
        """


class CustomNoiseAdder(NoiseAdder):
    """
    Instantiates a NoiseAdder using the provided `noise_filter`.
    """

    @property
    def noise_var(self, res=512):
        """
        Return noise variance.

        CustomNoiseAdder will estimate noise_var using the `noise_filter`.

        :param res: Resolution to use when evaluating noise filter, default 512.
        :returns: Noise variance estimated at `res`.
        """
        # Take mean of user provided _noise_filter, before the PowerFilter is applied.
        return np.mean(self._noise_filter.evaluate_grid(res))


class WhiteNoiseAdder(NoiseAdder):
    """
    A Xform that adds white noise, optionally passed through a Filter
    object, to all incoming images.
    """

    # TODO, check if we can change seed and/or why not.
    def __init__(self, var, seed=0):
        """
        Return a `WhiteNoiseAdder` instance from `var` and using `seed`.

        :param var: Target noise variance.
        :param seed: Optinally provide a random seed used to generate white noise.
        """

        self.signal_power = None  # Used with `from_snr`
        self.requires_signal_power = False  # Used with `from_snr`
        self.noise_var = var
        self.seed = seed
        # When we know the var, complete building the filter.
        if var is not None:
            self._build()
        else:
            # Otherwise, we will flag that we require signal power.
            # Assigning `signal_power` later will complete builing
            # filter.
            self.requires_signal_power = True

    def _build(self):
        """
        Builds underlying Filter for this NoiseAdder.
        """
        super().__init__(
            noise_filter=ScalarFilter(dim=2, value=self.noise_var), seed=self.seed
        )

    @classmethod
    def from_snr(cls, snr, signal_power=None, seed=0):
        """
        Generates a WhiteNoiseAdder configured to produce a target
        signal to noise ratio.

        When `signal_power` is not provided, `requires_signal_power`
        attribute will be set.  Consumers can check this attribute and
        set `signal_power` as required. Setting `signal_power` should
        then complete building the filter.

        :param snr: Desired signal to noise ratio of
            the returned source.
        :param signal_power: Optional, if the signal power is known.
        :param seed: Optionally provide a random seed used to generate white noise.
        """

        noise_adder = cls(var=None, seed=seed)
        # signal_power.setter will use `_snr` to compute the noise
        # variance.
        noise_adder._snr = snr

        # `signal_power.setter` should complete _build when provided
        # `signal_power` is not None
        noise_adder.signal_power = signal_power

        return noise_adder

    def __str__(self):
        return f"{self.__class__.__name__} with variance={self._noise_var}"

    def __repr__(self):
        return f"{self.__class__.__name__}(var={self._noise_var}, seed={self.seed})"

    @property
    def noise_var(self):
        """
        Returns noise variance.

        Note in this white noise case, noise variance is known,
        because the `WhiteNoiseAdder` was instantied with an explicit variance.
        """
        return self._noise_var

    @noise_var.setter
    def noise_var(self, v):
        self._noise_var = v

    @property
    def signal_power(self):
        return self._signal_power

    @signal_power.setter
    def signal_power(self, p):
        self._signal_power = p
        if p is not None:
            self.requires_signal_power = False
            self.noise_var = p / self._snr
            self._build()


class NoiseEstimator:
    """
    Noise Estimator base class.
    """

    def __init__(self, src, bgRadius=1, batchSize=512):
        """
        Any additional args/kwargs are passed on to the Source's 'images' method

        :param src: A Source object which can give us images on demand
        :param bgRadius: The radius of the disk whose complement is used to estimate the noise.
            Radius is relative proportion, where `1` represents
            the radius of disc inscribing a `(src.L, src.L)` image.
        :param batchSize:  The size of the batches in which to compute the variance estimate.
        """
        self.src = src
        self.dtype = self.src.dtype
        self.bgRadius = bgRadius
        self.batchSize = batchSize

        self.filter = self._create_filter()

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """
        raise NotImplementedError("Subclasses implement the `estimate` method.")


class WhiteNoiseEstimator(NoiseEstimator):
    """
    White Noise Estimator.
    """

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """
        # WhiteNoiseEstimator.filter is a ScalarFilter,
        #   so we only evaluate for the zero frequencies.
        return self.filter.evaluate(np.zeros((2, 1), dtype=self.dtype)).item()

    def _create_filter(self, noise_variance=None):
        """
        :param noise_variance: Noise variance of images
        :return: The estimated noise power spectral distribution (PSD) of the images in the form of a filter object.
        """
        if noise_variance is None:
            logger.info(f"Determining Noise variance in batches of {self.batchSize}")
            noise_variance = self._estimate_noise_variance()
            logger.info(f"Noise variance = {noise_variance}")
        return ScalarFilter(dim=2, value=noise_variance)

    def _estimate_noise_variance(self):
        """
        Any additional arguments/keyword-arguments are passed on to the Source's 'images' method
        :return: The estimated noise variance of the images in the Source used to create this estimator.
        TODO: How's this initial estimate of variance different from the 'estimate' method?
        """
        # Run estimate using saved parameters
        g2d = grid_2d(self.src.L, indexing="yx", dtype=self.dtype)
        mask = g2d["r"] >= self.bgRadius

        first_moment = 0
        second_moment = 0
        for i in trange(0, self.src.n, self.batchSize):
            images = self.src.images[i : i + self.batchSize].asnumpy()
            images_masked = images * mask

            _denominator = self.src.n * np.sum(mask)
            first_moment += np.sum(images_masked) / _denominator
            second_moment += np.sum(np.abs(images_masked**2)) / _denominator
        return second_moment - first_moment**2


class AnisotropicNoiseEstimator(NoiseEstimator):
    """
    Anisotropic White Noise Estimator.
    """

    def estimate(self):
        """
        :return: The estimated noise variance of the images.
        """

        # AnisotropicNoiseEstimator.filter is an ArrayFilter.
        #   We average the variance over all frequencies,

        return np.mean(self.filter.evaluate_grid(self.src.L))

    def _create_filter(self, noise_psd=None):
        """
        :param noise_psd: Noise PSD of images
        :return: The estimated noise power spectral distribution (PSD) of the images in the form of a filter object.
        """
        if noise_psd is None:
            noise_psd = self.estimate_noise_psd()
        return ArrayFilter(noise_psd)

    def estimate_noise_psd(self):
        """
        :return: The estimated noise variance of the images in the Source used to create this estimator.
        TODO: How's this initial estimate of variance different from the 'estimate' method?
        """
        # Run estimate using saved parameters
        g2d = grid_2d(self.src.L, indexing="yx", dtype=self.dtype)
        mask = g2d["r"] >= self.bgRadius

        mean_est = 0
        noise_psd_est = np.zeros((self.src.L, self.src.L)).astype(self.src.dtype)
        for i in trange(0, self.src.n, self.batchSize):
            images = self.src.images[i : i + self.batchSize].asnumpy()
            images_masked = images * mask

            _denominator = self.src.n * np.sum(mask)
            mean_est += np.sum(images_masked) / _denominator
            im_masked_f = xp.asnumpy(fft.centered_fft2(xp.asarray(images_masked)))
            noise_psd_est += np.sum(np.abs(im_masked_f**2), axis=0) / _denominator

        mid = self.src.L // 2
        noise_psd_est[mid, mid] -= mean_est**2

        return noise_psd_est



import logging

import numpy as np

from aspire.basis.basis_utils import all_besselj_zeros

logger = logging.getLogger(__name__)


class FBBasisMixin(object):
    """
    FBBasisMixin is a mixin implementing methods specific to Fourier-Bessel expansions,
    to be inherited by Fourier-Bessel subclasses of Basis.
    """

    def _calc_k_max(self):
        """
        Generate zeros of Bessel functions
        """
        # get upper_bound of zeros of Bessel functions
        upper_bound = min(self.ell_max + 1, 2 * self.nres + 1)

        # List of number of zeros
        n = []
        # List of zero values (each entry is an ndarray; all of possibly different lengths)
        zeros = []

        for ell in range(upper_bound):
            # for each ell, num_besselj_zeros returns the zeros of the
            # order ell Bessel function which are less than 2*pi*c*R = nres*pi/2,
            # the truncation rule for the Fourier-Bessel expansion
            if self.ndim == 2:
                bessel_order = ell
            elif self.ndim == 3:
                bessel_order = ell + 1 / 2
            _n, _zeros = all_besselj_zeros(bessel_order, self.nres * np.pi / 2)
            if _n == 0:
                break
            else:
                n.append(_n)
                zeros.append(_zeros)

        #  get maximum number of ell
        self.ell_max = len(n) - 1

        #  set the maximum of k for each ell
        self.k_max = np.array(n, dtype=int)

        # set the zeros for each ell
        # this is a ragged list of 1d ndarrays, where the i'th element is of size self.k_max[i]
        self.r0 = zeros



import logging
from abc import ABC, abstractmethod

import numpy as np
import ray
from ray.util.multiprocessing import Pool

from aspire import config
from aspire.classification.reddy_chatterji import reddy_chatterji_register
from aspire.image import Image, ImageStacker, MeanImageStacker
from aspire.utils import trange
from aspire.utils.coor_trans import grid_2d
from aspire.utils.multiprocessing import num_procs_suggestion

logger = logging.getLogger(__name__)


class Averager2D(ABC):
    """
    Base class for 2D Image Averaging methods.
    """

    def __init__(self, composite_basis, src, num_procs=None, dtype=None):
        """
        :param composite_basis:  Basis to be used during class average composition (eg FFB2D)
        :param src: Source of original images.
        :param num_procs: Number of processes to use.
            `None` will attempt computing a suggestion based on machine resources.
            Note some underlying code may already use threading.
        :param dtype: Numpy dtype to be used during alignment.
        """

        self.composite_basis = composite_basis
        self.src = src
        if dtype is None:
            if self.composite_basis:
                self.dtype = self.composite_basis.dtype
            elif self.src:
                self.dtype = self.src.dtype
            else:
                raise RuntimeError("You must supply a basis/src/dtype.")
        else:
            self.dtype = np.dtype(dtype)

        if num_procs is None:
            num_procs = num_procs_suggestion()
            # Only enable multiprocessing when several cores available
            if num_procs < 3:
                num_procs = 1
        elif not (isinstance(num_procs, int) and num_procs > 0):
            raise ValueError(
                f"num_procs should be a positive integer, passed {num_procs}."
            )
        self.num_procs = num_procs

        if self.src and self.dtype != self.src.dtype:
            logger.warning(
                f"{self.__class__.__name__} dtype {dtype}"
                f"does not match dtype of source {self.src.dtype}."
            )
        if self.composite_basis and self.dtype != self.composite_basis.dtype:
            logger.warning(
                f"{self.__class__.__name__} dtype {dtype}"
                f"does not match dtype of basis {self.composite_basis.dtype}."
            )

    @abstractmethod
    def average(
        self,
        classes,
        reflections,
        coefs=None,
    ):
        """
        Combines images using stacking in `self.composite_basis`.

        Subclasses should implement this.
        (Example EM algos use radically different averaging).

        Should return an Image source of synthetic class averages.

        :param classes: class indices, refering to src. (src.n, n_nbor).
        :param reflections: Bool representing whether to reflect image in `classes`.
            (n_clases, n_nbor)
        :param coefs: Optional basis coefs (could avoid recomputing).
            (src.n, coef_count)
        :return: Stack of synthetic class average images as Image instance.
        """

    def _cls_images(self, cls, src=None):
        """
        Util to return images as an array for class k (provided as
        array `cls` ), preserving the class/nbor order.  Note, Images
        will be a read-only view, copy if mutations required.

        :param cls: An iterable (0/1-D array or list) that holds the indices of images to align.
            In class averaging, this would be a class.
        :param src: Optionally override the src, for example, if you want to use a different
            source for a certain operation (ie alignment).
        """
        src = src or self.src
        return src.images[cls].asnumpy().astype(self.dtype, copy=False)


class AligningAverager2D(Averager2D):
    """
    Subclass supporting averagers which perfom an aligning stage.
    """

    def __init__(
        self,
        composite_basis,
        src,
        alignment_basis=None,
        image_stacker=None,
        num_procs=None,
        dtype=None,
    ):
        """
        :param composite_basis:  Basis to be used during class average composition (eg hi res Cartesian/FFB2D).
        :param src: Source of original images.
        :param alignment_basis: Optional, basis to be used only during alignment (eg FSPCA).
        :param image_stacker: Optional, provide a user defined `ImageStacker` instance,
            used during image stacking (averaging).  Defaults to MeanImageStacker.
        :param num_procs: Number of processes to use.
            `None` will attempt computing a suggestion based on machine resources.
            Note some underlying code may already use threading.
        :param dtype: Numpy dtype to be used during alignment.
        """

        super().__init__(
            composite_basis=composite_basis,
            src=src,
            num_procs=num_procs,
            dtype=dtype,
        )
        # If alignment_basis is None, use composite_basis
        self.alignment_basis = alignment_basis or self.composite_basis

        # If image_stacker is None, use mean
        self.image_stacker = image_stacker or MeanImageStacker()
        if not isinstance(self.image_stacker, ImageStacker):
            raise ValueError("`image_stacker` should be subclass of ImageStacker.")

        if not hasattr(self.composite_basis, "rotate"):
            raise RuntimeError(
                f"{self.__class__.__name__}'s composite_basis {self.composite_basis} must provide a `rotate` method."
            )
        if not hasattr(self.composite_basis, "shift"):
            raise RuntimeError(
                f"{self.__class__.__name__}'s composite_basis {self.composite_basis} must provide a `shift` method."
            )

    @abstractmethod
    def align(self, classes, reflections, basis_coefficients):
        """
        During this process `rotations`, `reflections`, `shifts` and
        `correlations` properties will be computed for aligners.

        `rotations` is an (src.n, n_nbor) array of angles,
        which should represent the rotations needed to align images within
        that class. `rotations` is measured in CCW radians.

        `shifts` is None or an (src.n, n_nbor) array of 2D shifts
        which should represent the translation needed to best align the images
        within that class.

        `correlations` is an (src.n, n_nbor) array representing
        a correlation like measure between classified images and their base
        image (image index 0).

        Subclasses of should implement and extend this method.

        :param classes: (src.n, n_nbor) integer array of img indices.
        :param reflections: (src.n, n_nbor) bool array of corresponding reflections,
        :param basis_coefficients: (n_img, self.alignment_basis.count) basis coefficients,

        :returns: (rotations, shifts, correlations)
        """

    def average(
        self,
        classes,
        reflections,
        coefs=None,
    ):
        """
        This subclass assumes we get alignment details from `align` method. Otherwise. see Averager2D.average
        """

        classes = np.atleast_2d(classes)
        reflections = np.atleast_2d(reflections)

        self.rotations, self.shifts, self.correlations = self.align(
            classes, reflections, coefs
        )

        n_classes, n_nbor = classes.shape

        b_avgs = np.empty((n_classes, self.composite_basis.count), dtype=self.src.dtype)

        def _innerloop(i):
            # Get coefs in Composite_Basis if not provided as an argument.
            if coefs is None:
                # Retrieve relevant images directly from source.
                neighbors_imgs = Image(self._cls_images(classes[i]))

                # Do shifts
                if self.shifts is not None:
                    neighbors_imgs = neighbors_imgs.shift(self.shifts[i])

                neighbors_coefs = self.composite_basis.evaluate_t(neighbors_imgs)
            else:
                # Get the neighbors
                neighbors_ids = classes[i]
                neighbors_coefs = coefs[neighbors_ids]
                if self.shifts is not None:
                    neighbors_coefs = self.composite_basis.shift(
                        neighbors_coefs, self.shifts[i]
                    )

            # Rotate in composite_basis
            neighbors_coefs = self.composite_basis.rotate(
                neighbors_coefs, self.rotations[i], reflections[i]
            )

            # Averaging in composite_basis
            return self.image_stacker(neighbors_coefs)

        if self.num_procs <= 1:
            for i in trange(n_classes):
                b_avgs[i] = _innerloop(i)
        else:
            logger.info(
                f"Starting Pool({self.num_procs}) for {self.__class__.__name__}.average"
            )
            ray.init(
                _temp_dir=config["ray"]["temp_dir"].as_filename(),
                num_cpus=self.num_procs,
                num_gpus=0,
            )
            with Pool(self.num_procs) as p:
                results = p.map(_innerloop, range(n_classes))
            ray.shutdown()

            logger.info(f"Terminated Pool({self.num_procs}), unpacking results.")
            for i, result in enumerate(results):
                b_avgs[i] = result

        # Now we convert the averaged images from Basis to Cartesian.
        return self.composite_basis.evaluate(b_avgs)

    def _shift_search_grid(self, L, radius, roll_zero=False):
        """
        Returns two 1-D arrays representing the X and Y grid points in the defined
        shift search space (disc <= self.radius).

        :param radius: Disc radius in pixels
        :returns: Grid points as 2-tuple of vectors X,Y.
        """

        # We'll brute force all shifts in a grid.
        g = grid_2d(L, normalized=False)
        disc = g["r"] <= radius
        X, Y = g["x"][disc], g["y"][disc]

        # Optionally roll arrays so 0 is first.
        if roll_zero:
            zero_ind = np.argwhere(X * X + Y * Y == 0).flatten()[0]
            X, Y = np.roll(X, -zero_ind), np.roll(Y, -zero_ind)
            assert (X[0], Y[0]) == (0, 0), (radius, zero_ind, X, Y)

        return X, Y


class BFRAverager2D(AligningAverager2D):
    """
    This perfoms a Brute Force Rotational alignment.

    For each class,
        constructs n_angles rotations of all class members,
        and then identifies angle yielding largest correlation(dot).
    """

    def __init__(
        self,
        composite_basis,
        src,
        alignment_basis=None,
        n_angles=360,
        num_procs=None,
        dtype=None,
    ):
        """
        See AligningAverager2D, adds:

        :param n_angles: Number of brute force rotations to attempt, defaults 360.
        """
        super().__init__(
            composite_basis, src, alignment_basis, num_procs=num_procs, dtype=dtype
        )

        self.n_angles = n_angles
        self._base_image_shift = None

        if not hasattr(self.alignment_basis, "rotate"):
            raise RuntimeError(
                f"{self.__class__.__name__}'s alignment_basis {self.alignment_basis} must provide a `rotate` method."
            )

    def align(self, classes, reflections, basis_coefficients):
        """
        Performs the actual rotational alignment estimation,
        returning parameters needed for averaging.
        """

        # Admit simple case of single case alignment
        classes = np.atleast_2d(classes)
        reflections = np.atleast_2d(reflections)

        n_classes, n_nbor = classes.shape

        # Construct array of angles to brute force.
        test_angles = np.linspace(0, -2 * np.pi, self.n_angles, endpoint=False)

        # Instantiate matrices for results
        rotations = np.empty(classes.shape, dtype=self.dtype)
        correlations = np.empty(classes.shape, dtype=self.dtype)

        def _innerloop(k):
            _correlations = np.full((n_nbor, self.n_angles), fill_value=-np.inf)
            _correlations[0, 0] = 1  # Set this now so we can skip it in the loop.
            # Get the coefs for these neighbors
            if basis_coefficients is None:
                # Retrieve relevant images
                neighbors_imgs = self._cls_images(classes[k]).copy()

                # We optionally can shift the base image by `_base_image_shift`
                # Shift in real space to avoid extra conversions
                if self._base_image_shift is not None:
                    neighbors_imgs[0] = (
                        Image(neighbors_imgs[0])
                        .shift(self._base_image_shift)
                        .asnumpy()[0]
                    )

                # Evaluate_t into basis
                nbr_coef = self.composite_basis.evaluate_t(Image(neighbors_imgs))
            else:
                nbr_coef = basis_coefficients[classes[k]]

            norm_0 = np.linalg.norm(nbr_coef[0])
            for i, angle in enumerate(test_angles):
                # Rotate the set of neighbors by angle,
                rotated_nbrs = self.alignment_basis.rotate(
                    nbr_coef, angle, reflections[k]
                )

                # then store dot between class base image (0) and each nbor
                for j, nbor in enumerate(rotated_nbrs):
                    # Skip the base image.
                    if j == 0:
                        continue
                    norm_nbor = np.linalg.norm(nbor)
                    _correlations[j, i] = np.dot(nbr_coef[0], nbor) / (
                        norm_nbor * norm_0
                    )

            # Now find the index of the angle reporting highest correlation
            angle_idx = np.argmax(_correlations, axis=1)

            # Take the correlation corresponding to angle_idx
            _correlations = np.take_along_axis(
                _correlations, np.expand_dims(angle_idx, axis=1), axis=1
            ).reshape(n_nbor)

            return test_angles[angle_idx], _correlations

        if self.num_procs <= 1:
            for k in trange(n_classes):
                # Store angles and correlations for this class
                rotations[k], correlations[k] = _innerloop(k)

        else:
            logger.info(
                f"Starting Pool({self.num_procs}) for {self.__class__.__name__}.align"
            )
            ray.init(
                _temp_dir=config["ray"]["temp_dir"].as_filename(),
                num_cpus=self.num_procs,
                num_gpus=0,
            )
            with Pool(self.num_procs) as p:
                results = p.map(_innerloop, range(n_classes))
            ray.shutdown()

            logger.info(f"Terminated Pool({self.num_procs}), unpacking results.")
            for k, result in enumerate(results):
                rotations[k], correlations[k] = result

        return rotations, None, correlations


class BFSRAverager2D(BFRAverager2D):
    """
    This perfoms a Brute Force Shift and Rotational alignment.
    It is potentially expensive to brute force this search space.

    For each pair of x_shifts and y_shifts,
       Perform BFR

    Return the rotation and shift yielding the best results.
    """

    def __init__(
        self,
        composite_basis,
        src,
        alignment_basis=None,
        n_angles=360,
        radius=None,
        num_procs=None,
        dtype=None,
    ):
        """
        See AligningAverager2D and BFRAverager2D, adds: radius

        :params n_angles: Number of brute force rotations to attempt, defaults 360.
        :param radius: Brute force translation search radius.
            Defaults to src.L//16.
        """
        super().__init__(
            composite_basis,
            src,
            alignment_basis,
            n_angles,
            num_procs=num_procs,
            dtype=dtype,
        )

        self.radius = radius if radius is not None else src.L // 16

        # Each shift will require calling the parent BFRAverager2D.align
        self._bfr_align = super().align

        if not hasattr(self.alignment_basis, "shift"):
            raise RuntimeError(
                f"{self.__class__.__name__}'s alignment_basis {self.alignment_basis} must provide a `shift` method."
            )

    def align(self, classes, reflections, basis_coefficients):
        """
        See `AligningAverager2D.align`
        """

        # Admit simple case of single case alignment
        classes = np.atleast_2d(classes)
        reflections = np.atleast_2d(reflections)

        n_classes = classes.shape[0]

        # Create a search grid and force initial pair to (0,0)
        # This is done primarily in case of a tie later, we would take unshifted.
        x_shifts, y_shifts = self._shift_search_grid(
            self.src.L, self.radius, roll_zero=True
        )

        # These arrays will incrementally store our best alignment.
        rotations = np.empty(classes.shape, dtype=self.dtype)
        correlations = np.ones(classes.shape, dtype=self.dtype) * -np.inf
        shifts = np.empty((*classes.shape, 2), dtype=int)

        # We want to maintain the original coefs for the base images,
        #  because we will mutate them with shifts in the loop.
        if basis_coefficients is None:
            original_coef = self.composite_basis.evaluate_t(
                Image(self._cls_images(classes[:, 0], src=self.src))
            )
        else:
            original_coef = basis_coefficients[classes[:, 0], :].copy()

        # Sanity check the original_coef shape
        assert original_coef.shape == (n_classes, self.alignment_basis.count)

        # Loop over shift search space, updating best result
        for x, y in zip(x_shifts, y_shifts):
            shift = np.array([x, y], dtype=int)
            logger.debug(f"Computing rotational alignment after shift ({x},{y}).")

            # Shift the coef representing the first (base) entry in each class
            #   by the negation of the shift
            # Shifting one image is more efficient than shifting every neighbor
            if basis_coefficients is not None:
                basis_coefficients[classes[:, 0], :] = self.alignment_basis.shift(
                    original_coef, -shift
                )
            else:
                # Store the latest shift so that super class can access it.
                # This allows us to retrieve and shift coefficients on the fly,
                #   instead of storing them all.
                self._base_image_shift = -shift

            _rotations, _, _correlations = self._bfr_align(
                classes, reflections, basis_coefficients
            )

            # Each class-neighbor pair may have a best shift-rot from a different shift.
            # Test and update
            improved_indices = _correlations > correlations
            rotations[improved_indices] = _rotations[improved_indices]
            correlations[improved_indices] = _correlations[improved_indices]
            shifts[improved_indices] = shift

            # Cleanup/Restore unshifted base coefs
            if basis_coefficients is None:
                # Reset this flag
                self._base_image_shift = None
            else:
                basis_coefficients[classes[:, 0], :] = original_coef

            if (x, y) == (0, 0):
                logger.debug("Initial rotational alignment complete (shift (0,0))")
                assert np.sum(improved_indices) == np.size(
                    classes
                ), f"{np.sum(improved_indices)} =?= {np.size(classes)}"
            else:
                logger.debug(
                    f"Shift ({x},{y}) complete. Improved {np.sum(improved_indices)} alignments."
                )

        return rotations, shifts, correlations


class ReddyChatterjiAverager2D(AligningAverager2D):
    """
    Attempts rotational estimation using Reddy Chatterji log polar Fourier cross correlation.
    Then attempts shift (translational) estimation using cross correlation.

    When averaging, performs rotations then shifts.

    Note, it may be possible to iterate this algorithm...

    Adopted from Reddy Chatterji  (1996)
    An FFT-Based Technique for Translation,
    Rotation, and Scale-Invariant Image Registration
    IEEE TRANSACTIONS ON IMAGE PROCESSING, VOL. 5, NO. 8, AUGUST 1996

    This method intentionally does not use any of ASPIRE's basis
    so that it may be used as a reference for more ASPIRE approaches.
    """

    def __init__(
        self,
        composite_basis,
        src,
        alignment_src=None,
        num_procs=None,
        dtype=None,
    ):
        """
        :param composite_basis:  Basis to be used during class average composition.
        :param src: Source of original images.
        :param alignment_src: Optional, source to be used during class average alignment.
            Must be the same resolution as `src`.
        :param num_procs: Number of processes to use.
            `None` will attempt computing a suggestion based on machine resources.
            Note some underlying code may already use threading.
        :param dtype: Numpy dtype to be used during alignment.
        """

        self.alignment_src = alignment_src or src

        # TODO, for accomodating different resolutions we minimally need to adapt shifting.
        # Outside of scope right now, but would make a nice PR later.
        if self.alignment_src.L != src.L:
            raise RuntimeError("Currently `alignment_src.L` must equal `src.L`")
        if self.alignment_src.dtype != src.dtype:
            raise RuntimeError("Currently `alignment_src.dtype` must equal `src.dtype`")

        self.mask = grid_2d(src.L, normalized=False)["r"] < src.L // 2

        super().__init__(
            composite_basis, src, composite_basis, num_procs=num_procs, dtype=dtype
        )

    def align(self, classes, reflections, basis_coefficients):
        """
        Performs the actual rotational alignment estimation,
        returning parameters needed for averaging.
        """

        # Admit simple case of single case alignment
        classes = np.atleast_2d(classes)
        reflections = np.atleast_2d(reflections)

        n_classes = classes.shape[0]

        # Instantiate matrices for results
        rotations = np.zeros(classes.shape, dtype=self.dtype)
        correlations = np.zeros(classes.shape, dtype=self.dtype)
        shifts = np.zeros((*classes.shape, 2), dtype=int)

        def _innerloop(k):
            # Get the array of images for this class, using the `alignment_src`.
            images_k = self._cls_images(classes[k], src=self.alignment_src)
            return reddy_chatterji_register(
                images_k,
                reflections[k],
                mask=self.mask,
                do_cross_corr_translations=True,
                dtype=self.dtype,
            )

        if self.num_procs <= 1:
            for k in trange(n_classes):
                rotations[k], shifts[k], correlations[k] = _innerloop(k)

        else:
            logger.info(
                f"Starting Pool({self.num_procs}) for {self.__class__.__name__}.align"
            )
            ray.init(
                _temp_dir=config["ray"]["temp_dir"].as_filename(),
                num_cpus=self.num_procs,
                num_gpus=0,
            )
            with Pool(self.num_procs) as p:
                results = p.map(_innerloop, range(n_classes))
            ray.shutdown()

            logger.info(f"Terminated Pool({self.num_procs}), unpacking results.")
            for k, result in enumerate(results):
                rotations[k], shifts[k], correlations[k] = result

        return rotations, shifts, correlations

    def average(
        self,
        classes,
        reflections,
        coefs=None,
    ):
        """
        This averages classes performing rotations then shifts.
        Otherwise is similar to `AligningAverager2D.average`.
        """

        self.rotations, self.shifts, self.correlations = self.align(
            classes, reflections, coefs
        )

        n_classes, n_nbor = classes.shape

        b_avgs = np.empty((n_classes, self.composite_basis.count), dtype=self.src.dtype)

        def _innerloop(i):
            # Get coefs in Composite_Basis if not provided as an argument.
            if coefs is None:
                # Retrieve relevant images directly from source.
                neighbors_imgs = Image(self._cls_images(classes[i]))
                neighbors_coefs = self.composite_basis.evaluate_t(neighbors_imgs)
            else:
                # Get the neighbors
                neighbors_ids = classes[i]
                neighbors_coefs = coefs[neighbors_ids]

            # Rotate in composite_basis
            neighbors_coefs = self.composite_basis.rotate(
                neighbors_coefs, self.rotations[i], reflections[i]
            )

            # Note shifts are after rotation for this approach!
            if self.shifts is not None:
                neighbors_coefs = self.composite_basis.shift(
                    neighbors_coefs, self.shifts[i]
                )

            # Averaging in composite_basis
            return self.image_stacker(neighbors_coefs)

        if self.num_procs <= 1:
            for i in trange(n_classes):
                b_avgs[i] = _innerloop(i)
        else:
            logger.info(
                f"Starting Pool({self.num_procs}) for {self.__class__.__name__}.average"
            )
            ray.init(
                _temp_dir=config["ray"]["temp_dir"].as_filename(),
                num_cpus=self.num_procs,
                num_gpus=0,
            )
            with Pool(self.num_procs) as p:
                results = p.map(_innerloop, range(n_classes))
            ray.shutdown()

            logger.info(f"Terminated Pool({self.num_procs}), unpacking results.")
            for i, result in enumerate(results):
                b_avgs[i] = result

        # Now we convert the averaged images from Basis to Cartesian.
        return self.composite_basis.evaluate(b_avgs)


class BFSReddyChatterjiAverager2D(ReddyChatterjiAverager2D):
    """
    Brute Force Shifts (Translations) - ReddyChatterji (Log-Polar) Rotations

    For each shift within `radius`, attempts rotational match using ReddyChatterji.
    When averaging, performs shift before rotations,

    Adopted from Reddy Chatterji  (1996)
    An FFT-Based Technique for Translation,
    Rotation, and Scale-Invariant Image Registration
    IEEE TRANSACTIONS ON IMAGE PROCESSING, VOL. 5, NO. 8, AUGUST 1996

    This method intentionally does not use any of ASPIRE's basis
    so that it may be used as a reference for more ASPIRE approaches.
    """

    def __init__(
        self,
        composite_basis,
        src,
        alignment_src=None,
        radius=None,
        num_procs=None,
        dtype=None,
    ):
        """
        :param alignment_basis: Basis to be used during alignment.
            For current implementation of ReddyChatterjiAverager2D this should be `None`.
            Instead see `alignment_src`.
        :param src: Source of original images.
        :param composite_basis:  Basis to be used during class average composition.
        :param alignment_src: Optional, source to be used during class average alignment.
            Must be the same resolution as `src`.
        :param radius: Brute force translation search radius.
            Defaults to src.L//8.
        :param dtype: Numpy dtype to be used during alignment.
        :param num_procs: Number of processes to use.
            `None` will attempt computing a suggestion based on machine resources.
            Note some underlying code may already use threading.
        :param dtype: Numpy dtype to be used during alignment.
        """

        super().__init__(
            composite_basis,
            src,
            alignment_src,
            num_procs=num_procs,
            dtype=dtype,
        )

        # Assign search radius
        self.radius = radius if radius is not None else src.L // 8

    def align(self, classes, reflections, basis_coefficients):
        """
        Performs the actual rotational alignment estimation,
        returning parameters needed for averaging.
        """

        # Admit simple case of single case alignment
        classes = np.atleast_2d(classes)
        reflections = np.atleast_2d(reflections)

        n_classes, n_nbor = classes.shape

        # Instantiate matrices for inner loop, and best results.
        rotations = np.zeros(classes.shape, dtype=self.dtype)
        correlations = np.ones(classes.shape, dtype=self.dtype) * -np.inf
        shifts = np.zeros((*classes.shape, 2), dtype=int)

        X, Y = self._shift_search_grid(self.alignment_src.L, self.radius)

        def _innerloop(k):
            unshifted_images = self._cls_images(classes[k])
            # Instantiate matrices for inner loop, and best results.
            _rotations = np.zeros(classes.shape[1:], dtype=self.dtype)
            _correlations = np.ones(classes.shape[1:], dtype=self.dtype) * -np.inf
            _shifts = np.zeros((*classes.shape[1:], 2), dtype=int)

            for xs, ys in zip(X, Y):
                s = np.array([xs, ys])
                # Get the array of images for this class

                # Note we mutate `images` here with shifting,
                #   then later reddy_chatterji_register
                images = unshifted_images.copy()
                # Don't shift the base image
                images[1:] = Image(unshifted_images[1:]).shift(s).asnumpy()

                # returned shifts ignored since we are forcing shift of `s` above
                __rotations, _, __correlations = reddy_chatterji_register(
                    images,
                    reflections[k],
                    mask=self.mask,
                    do_cross_corr_translations=False,  # When forcing s, we skip cross corr translations
                    dtype=self.dtype,
                )

                # Where corr has improved
                #  update our rolling best results with this loop.
                improved = __correlations > _correlations
                _correlations = np.where(improved, __correlations, _correlations)
                _rotations = np.where(improved, __rotations, _rotations)
                _shifts = np.where(improved[..., np.newaxis], s, _shifts)
                logger.debug(f"Shift {s} has improved {np.sum(improved)} results")

            return _rotations, _shifts, _correlations

        if self.num_procs <= 1:
            for k in trange(n_classes):
                rotations[k], shifts[k], correlations[k] = _innerloop(k)
        else:
            logger.info(
                f"Starting Pool({self.num_procs}) for {self.__class__.__name__}.align"
            )
            ray.init(
                _temp_dir=config["ray"]["temp_dir"].as_filename(),
                num_cpus=self.num_procs,
                num_gpus=0,
            )
            with Pool(self.num_procs) as p:
                results = p.map(_innerloop, range(n_classes))
            ray.shutdown()

            logger.info(f"Terminated Pool({self.num_procs}), unpacking results.")
            for k, result in enumerate(results):
                rotations[k], shifts[k], correlations[k] = result

        return rotations, shifts, correlations

    def average(
        self,
        classes,
        reflections,
        coefs=None,
    ):
        """
        See Averager2D.average.
        """
        # ReddyChatterjiAverager2D does rotations then shifts.
        # For brute force, we'd like shifts then rotations,
        #   as is done in general in AligningAverager2D
        return AligningAverager2D.average(self, classes, reflections, coefs)


class EMAverager2D(Averager2D):
    """
    Citation needed.
    """


class FTKAverager2D(Averager2D):
    """
    Factorization of the translation kernel for fast rigid image alignment.
    Rangan, A.V., Spivak, M., Anden, J., & Barnett, A.H. (2019).
    """



from .volume import Volume, qr_vols_forward, rotated_grids, rotated_grids_3d

from .volume_synthesis import (  # isort:skip
    TSymmetricVolume,
    OSymmetricVolume,
    CnSymmetricVolume,
    DnSymmetricVolume,
    AsymmetricVolume,
    LegacyVolume,
)



import logging

import numpy as np

from aspire.image import Image
from aspire.source import ImageSource

logger = logging.getLogger(__name__)


class DenoisedImageSource(ImageSource):
    """
    Define a derived ImageSource class to perform operations for denoised 2D images
    """

    def __init__(self, src, denoiser, batch_size=512):
        """
        Initialize a denoised ImageSource object from original ImageSource of noisy images

        :param src: Original ImageSource object storing noisy images
        :param denoiser: A Denoiser object for specifying a method for denoising
        :param batch_size: Batch size for loading denoised images.
        """

        super().__init__(src.L, src.n, dtype=src.dtype, metadata=src._metadata.copy())
        self._im = None
        self.denoiser = denoiser
        self.batch_size = batch_size

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _images(self, indices):
        """
        Internal function to return a set of images after denoising, when accessed via the
        `ImageSource.images` property.
        :param indices: The indices of images to return as a 1-D NumPy array.
        :return: an `Image` object after denoisng.
        """
        # check for cached images first
        if self._cached_im is not None:
            logger.info("Loading images from cache")
            return self.generation_pipeline.forward(
                self._cached_im[indices, :, :], indices
            )

        # start and end (and indices) refer to the indices in the DenoisedImageSource
        # that are being denoised and returned in batches
        start = indices.min()
        end = indices.max()

        nimgs = len(indices)
        im = np.empty((nimgs, self.L, self.L), self.dtype)

        # If we request less than a whole batch, don't crash
        batch_size = min(nimgs, self.batch_size)

        logger.info(f"Loading {nimgs} images complete")
        for batch_start in range(start, end + 1, batch_size):
            imgs_denoised = self.denoiser.images(batch_start, batch_size)
            batch_end = min(batch_start + batch_size, end + 1)
            # we subtract start here to correct for any offset in the indices
            im[batch_start - start : batch_end - start] = imgs_denoised.asnumpy()

        # Finally, apply transforms to resulting Image
        return self.generation_pipeline.forward(Image(im), indices)



"""
Miscellaneous utilities for common data type operations.
"""

import logging

import numpy as np

logger = logging.getLogger(__name__)


def real_type(complextype):
    """
    Get Numpy real type from corresponding complex type

    :param complextype: Numpy complex type
    :return realtype: Numpy real type
    """
    complextype = np.dtype(complextype)
    realtype = None
    if complextype == np.complex64:
        realtype = np.float32
    elif complextype == np.complex128:
        realtype = np.float64
    elif complextype in (np.float32, np.float64):
        realtype = complextype
    else:
        msg = f"Corresponding real type is not defined for {complextype}."
        logger.error(msg)
        raise TypeError(msg)

    return realtype


def complex_type(realtype):
    """
    Get Numpy complex type from corresponding real type

    :param realtype: Numpy real type
    :return complextype: Numpy complex type
    """
    realtype = np.dtype(realtype)
    complextype = None
    if realtype == np.float32:
        complextype = np.complex64
    elif realtype == np.float64:
        complextype = np.complex128
    elif realtype in (np.complex64, np.complex128):
        complextype = realtype
    else:
        msg = f"Corresponding complex type is not defined for {realtype}."
        logger.error(msg)
        raise TypeError(msg)

    return complextype


def utest_tolerance(dtype):
    """
    Return ASPIRE tolerance for unit tests based on `dtype`.
    """

    if dtype == np.float64:
        # Use default np.allclose atol
        tol = 1e-8
    elif dtype == np.float32:
        tol = 1e-5
    else:
        msg = f"utest_tolerance not implemented for dtype: {dtype}."
        logger.error(msg)
        raise TypeError(msg)

    return tol



import logging
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from math import floor

import mrcfile
import numpy as np

from aspire.image import Image
from aspire.operators import CTFFilter, IdentityFilter
from aspire.source.image import ImageSource
from aspire.storage import StarFile
from aspire.utils import RelionStarFile

logger = logging.getLogger(__name__)


class CoordinateSource(ImageSource, ABC):
    """
    Base class defining common methods for data sources consisting of full
    micrographs coupled with files specifying the locations of picked
    particles in each micrograph.

    Broadly, there are two ways this information is represented. Sometimes each
    coordinate is simply the (X,Y) center location of the picked particle. This
    is sometimes stored in a `.coord` text file, and sometimes in a STAR file.
    The particle box is then computed using an externally supplied box size.
    These sources may be loaded via the `CentersCoordinateSource` class for
    both filetypes.

    Other formats adhere to the box file specification, which
    specifies a particle via four numbers:
    (lower left X coordinate, lower left Y coordinate, X size, Y size).
    These can be loaded via the `BoxesCoordinateSource` class.

    Regardless of source, the coordinates of each particle are represented
    internally in the box format.

    Particle information is extracted from the micrographs and coordinate files
    and put into a common data structure (self.particles).

    The `_images()` method, called via `ImageSource.images()` crops
    the particle images out of the micrograph and returns them as a stack.
    This also allows the CoordinateSource to be saved to an `.mrcs` stack.
    """

    def __init__(self, files, particle_size, max_rows, B):
        mrc_paths, coord_paths = [f[0] for f in files], [f[1] for f in files]
        # the particle_size parameter is the *user-specified* argument
        # and is used in self._populate_particles
        # it may be None in the case of an BoxesCoordinateSource
        self.particle_size = particle_size

        # keep this list to identify micrograph paths by index rather than
        # storing many copies of the same string
        self.mrc_paths = mrc_paths
        self.num_micrographs = len(self.mrc_paths)

        # The internal representation of micrographs and their picked coords
        # is a list of tuples (index of micrograph, coordinate), where
        # the coordinate is a list of the form:
        # [lower left X, lower left Y, size X, size Y].
        # The micrograph's filepath can be recovered from self.mrc_paths
        self.particles = []
        self._populate_particles(coord_paths)

        # Read shapes of all micrographs
        self.mrc_shapes = self._get_mrc_shapes()

        # get first mrc and coordinate file to report some data
        first_mrc_index, first_coord = self.particles[0]
        first_mrc = self.mrc_paths[first_mrc_index]

        with mrcfile.open(first_mrc) as mrc:
            # get dtype from first micrograph
            mode = int(mrc.header.mode)
            dtypes = {0: "int8", 1: "int16", 2: "float32", 6: "uint16"}
            assert (
                mode in dtypes
            ), f"Only modes={list(dtypes.keys())} in MRC files are supported for now."
            dtype = dtypes[mode]

        # look at first coord to get the particle size
        # this was either provided by the user or read from a .box file
        # here we're checking the final coordinate of the first particle
        # which is the Y-size of the box (the same as the X-size)
        # this is not the same as the argument particle_size
        # which can be None
        L = first_coord[3]

        logger.info(f"Particle size = {L}x{L}")
        self._original_resolution = L

        # total micrographs and particles represented by source (info)
        logger.info(
            f"{self.__class__.__name__} from {os.path.dirname(self.mrc_paths[0])} contains {self.num_micrographs} micrographs, {len(self.particles)} picked particles."
        )
        # report different mrc shapes
        logger.info(f"Micrographs have the following shapes: {*set(self.mrc_shapes),}")

        # remove particles whose boxes do not fit at given particle_size
        # and get number removed
        boundary_removed = self._exclude_boundary_particles()

        # total particles we can load given particle_size (info)
        if boundary_removed > 0:
            logger.info(
                f"{boundary_removed} particles did not fit into micrograph dimensions at particle size {L}, so were excluded"
            )
            logger.info(
                f"Maximum number of particles at this particle size is {len(self.particles)}."
            )

        # if max_rows is specified, only return up to max_rows many
        # (after excluding boundary particles)
        if max_rows:
            max_rows = min(max_rows, len(self.particles))
            self.particles = self.particles[:max_rows]

        # final number of particles in *this* source
        n = len(self.particles)

        # total particles loaded (specific to this instance)
        logger.info(f"CoordinateSource object contains {n} particles.")

        ImageSource.__init__(self, L=L, n=n, dtype=dtype)

        # map mrc indices to particle indices
        # i'th element contains a list of particle indices corresponding to i'th mrc
        self.mrc_index_to_particles = []
        for mrc_idx in range(self.num_micrographs):
            self.mrc_index_to_particles.append(
                [
                    particle_idx
                    for particle_idx, particle in enumerate(self.particles)
                    if particle[0] == mrc_idx
                ]
            )

        # CTF envelope decay factor
        self.B = B
        # set CTF metadata to defaults
        # this can be updated with import_ctf()
        self.filter_indices = np.zeros(self.n, dtype=int)
        self.unique_filters = [IdentityFilter()]
        self.set_metadata("__filter_indices", np.zeros(self.n, dtype=int))

        # populate __mrc_filename and __mrc_index
        for mrc_index, particle_indices in enumerate(self.mrc_index_to_particles):
            self.set_metadata(
                "__mrc_index",
                mrc_index,
                particle_indices,
            )
            self.set_metadata(
                "__mrc_filepath", self.mrc_paths[mrc_index], particle_indices
            )

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _populate_particles(self, coord_paths):
        """
        All subclasses create mrc_paths and coord_paths lists and pass them to
        this method.
        """
        for i in range(self.num_micrographs):
            # read in all coordinates for the given mrc using subclass's
            # method of reading the corresponding coord file
            self.particles += [
                (i, coord) for coord in self._coords_list_from_file(coord_paths[i])
            ]

    @abstractmethod
    def _coords_list_from_file(self, coord_file):
        """
        Given a coordinate file, convert the coordinates into box format, i.e. a
        list of the form [lower left x, lower left y, x size, y size].
        Subclasses implement according to the details of the files they read.
        """

    @staticmethod
    def _box_coord_from_center(center, particle_size):
        """
        Convert a list `[x,y]` representing a particle center
        to a list
        `[lower left x, lower left y, particle_size, particle_size]`
        representing the box around the particle in box format.
        :param center: a list of length two representing a center
        :param particle_size: the size of the box around the particle
        """
        # subtract off floor(particle size/2) from center coords
        r = floor(particle_size / 2)
        x, y = center[:2]
        return [
            # centers may be represented as floats in STAR files
            # chop off the non integer part to account for this
            int(x) - r,
            int(y) - r,
            particle_size,
            particle_size,
        ]

    @staticmethod
    def _center_from_box_coord(box_coord):
        """
        Convert a list
        `[lower left x, lower left y, particle_size, particle_size]`
        representing a particle in the box format to a list
        `[x, y]` representing the particle center.
        :param box_coord: a list of length 4 representing the particle box
        """
        # Get lower left corner x and y coordinates
        # and particle size from the box coordinate
        llx, lly, particle_size = box_coord[:3]
        # turn lower left corner coordinate into center coordinate
        r = floor(particle_size / 2)
        return [llx + r, lly + r]

    def _coords_list_from_star(self, star_file):
        """
        Given a Relion STAR coordinate file (generally containing particle centers)
        return a list of coordinates in box format.
        :param star_file: A path to a STAR file containing particle centers
        """
        data_block = StarFile(star_file).get_block_by_index(0)
        coords = list(zip(data_block["_rlnCoordinateX"], data_block["_rlnCoordinateY"]))
        coords = [(float(x), float(y)) for x, y in coords]
        return [
            self._box_coord_from_center(coord, self.particle_size) for coord in coords
        ]

    def _populate_local_metadata(self):
        """
        Called during ImageSource.save(), populates metadata columns specific to
            `CoordinateSource` when saving to STAR file.
        :return: A list of the names of the columns added.
        """
        # Insert stored particle coordinates (centers) into metadata
        self.set_metadata(
            "_rlnCoordinateX",
            [
                self._center_from_box_coord(particle[1])[0]
                for particle in self.particles
            ],
        )
        self.set_metadata(
            "_rlnCoordinateY",
            [
                self._center_from_box_coord(particle[1])[1]
                for particle in self.particles
            ],
        )
        return ["_rlnCoordinateX", "_rlnCoordinateY"]

    def _exclude_boundary_particles(self):
        """
        Remove particles boxes which do not fit in the micrograph
        with the given `particle_size`.
        :return: Number of particles removed
        """
        out_of_range = []
        for i, particle in enumerate(self.particles):
            start_x, start_y, size_x, size_y = particle[1]
            # get shape of corresponding micrograph
            mrc_index = particle[0]
            mrc_shape = self.mrc_shapes[mrc_index]
            if (
                start_x < 0
                or start_y < 0
                or (start_x + size_x >= mrc_shape[1])
                or (start_y + size_y >= mrc_shape[0])
            ):
                out_of_range.append(i)

        # out_of_range stores the indices of the particles in the
        # unmodified coord_list that we must remove.
        # If we pop these indices of _coord list going forward, the
        # following indices will be shifted. Thus we pop in reverse, since
        # the indices prior to each removed index are unchanged
        for j in reversed(out_of_range):
            self.particles.pop(j)

        return len(out_of_range)

    def _get_mrc_shapes(self):
        """
        Iterate through self.mrc_shapes and read the dimensions of each micrograph
        :return mrc_shapes: A list of tuples representing the corresponding shapes
        """
        mrc_shapes = [(0, 0) for mrc in self.mrc_paths]
        for i, mrc in enumerate(self.mrc_paths):
            with mrcfile.open(mrc) as mrc_file:
                shape = mrc_file.data.shape
                if len(shape) != 2:
                    raise ValueError(
                        f"Shape of mrc file is {shape} but expected shape of size 2."
                        "Is this a stack of unaligned micrographs?"
                    )
                mrc_shapes[i] = shape

        return mrc_shapes

    def import_aspire_ctf(self, ctf):
        """
        Import CTF information from STAR file(s) generated by ASPIRE's CTF Estimator.

        :param ctf: A path or iterable of paths to STAR files from ASPIRE's CTF Estimator.
            Note that number of files provided must match number of micrographs in this
            `CoordinateSource`.
        """
        if not isinstance(ctf, Iterable):
            ctf = [ctf]
        if not len(ctf) == self.num_micrographs:
            raise ValueError(
                "Number of CTF STAR files must match number of micrographs."
            )

        # merge dicts from CTF files
        data_blocks = defaultdict(list)
        for f in ctf:
            # ASPIRE's CTF Estimator produces legacy (=< 3.0) STAR files containing one row
            star = RelionStarFile(f)
            data_block = star.data_block
            for k, v in data_block.items():
                data_blocks[k].append(v)

        self._extract_ctf(data_blocks)

    def import_relion_ctf(self, ctf):
        """
        Import CTF information Relion micrograph STAR files containing CTF information.

        :param ctf: Path to a Relion micrograph STAR file containing CTF information.
            Note that number of files provided must match number of micrographs in this
            `CoordinateSource`.
        """
        data_block = RelionStarFile(ctf).get_merged_data_block()

        # data_block is a dict containing the micrographs
        if not len(list(data_block.values())[0]) == self.num_micrographs:
            raise ValueError(
                f"{ctf} has CTF information for {len(data_block)}",
                f" micrographs but this source has {self.num_micrographs} micrographs.",
            )

        self._extract_ctf(data_block)

    def _extract_ctf(self, data_block):
        """
        Receives a dict containing micrograph CTF information, and populates
            the Source's CTF Filters, filter indices, and metadata.
        """
        # required CTF params excluding pixel size
        CTF_params = [
            "_rlnVoltage",
            "_rlnDefocusU",
            "_rlnDefocusV",
            "_rlnDefocusAngle",
            "_rlnSphericalAberration",
            "_rlnAmplitudeContrast",
            "_rlnMicrographPixelSize",
        ]

        # get unique ctfs from the data block
        # i'th entry of `indices` contains the index of `filter_params` with corresponding CTF params
        ctf_data = np.stack([data_block[c] for c in CTF_params]).astype(self.dtype).T
        filter_params, indices = np.unique(
            ctf_data,
            return_inverse=True,
            axis=0,
        )

        # convert defocus_ang from degrees to radians
        filter_params[:, 3] *= np.pi / 180.0

        # construct filters
        self.unique_filters = [
            CTFFilter(
                pixel_size=filter_params[i, 6],
                voltage=filter_params[i, 0],
                defocus_u=filter_params[i, 1],
                defocus_v=filter_params[i, 2],
                defocus_ang=filter_params[i, 3],
                Cs=filter_params[i, 4],
                alpha=filter_params[i, 5],
                B=self.B,
            )
            for i in range(len(filter_params))
        ]

        # set metadata
        for mrc_idx, filter_index in enumerate(indices):
            particle_indices_this_mrc = self.mrc_index_to_particles[mrc_idx]
            num_particles_this_mrc = len(particle_indices_this_mrc)
            self.set_metadata(
                CTF_params,
                np.vstack((filter_params[filter_index],) * num_particles_this_mrc),
                particle_indices_this_mrc,
            )
            self.set_metadata(
                "__filter_indices", filter_index, particle_indices_this_mrc
            )

    @staticmethod
    def _crop_micrograph(data, coord):
        """
        Crops a particle box defined by `coord` out of `data`.
        According to MRC 2014 convention, the origin represents the bottom-left
        corner of the image.
        :param data: A 2D numpy array representing a micrograph
        :param coord: A list of integers: (lower left X, lower left Y, X, Y)
        """
        start_x, start_y, size_x, size_y = coord
        return data[start_y : start_y + size_y, start_x : start_x + size_x]

    def _images(self, indices):
        """
        Given a range or selection of indices, returns an Image stack
        when accessed via the `ImageSource.images` property.
        Note that the indices refer to the order
        of the particles loaded in this *specific* CoordinateSource. This may
        not correspond to the particles in the original source on disk, if some
        particles were excluded due to their box not fitting into the mrc
        dimensions. Thus, the exact particles returned are a function of the
        `particle_size`.
        :param indices: A 1-D NumPy array of integer indices.
        :return: An `Image` object.
        """
        # check for cached images first
        if self._cached_im is not None:
            logger.info("Loading images from cache")
            return self.generation_pipeline.forward(
                self._cached_im[indices, :, :], indices
            )

        logger.info(f"Loading {len(indices)} images from micrographs")

        selected_particles = [self.particles[i] for i in indices]
        # initialize empty array to hold particle stack
        im = np.empty(
            (len(indices), self._original_resolution, self._original_resolution),
            dtype=self.dtype,
        )

        # group particles by micrograph in order to
        # only open each one once
        grouped = defaultdict(list)
        # this creates a dict of the form
        # { mrc_index : list of coords in that mrc, with relative order preserved }
        for mrc_index, coord in selected_particles:
            grouped[mrc_index].append(coord)

        # this loops over the micrographs
        # now that the particles have been grouped by
        # their origin micrograph
        for mrc_index, coord_list in grouped.items():
            # get explicit filepath from cached list
            fp = self.mrc_paths[mrc_index]
            with mrcfile.open(fp) as mrc_in:
                arr = mrc_in.data
            # create iterable of the coordinates in this mrc
            # we don't need to worry about exhausting this iter
            # because we know it contains the exact number of particles
            # selected from this micrograph.
            # the next()  method on this iterable will give us the next
            # particle from this micrograph to appear in selected_particles
            coord = iter(coord_list)
            # iterate through selected particles
            # we are potentially populating the particles
            # out of order, to optimize the slower operation
            # of opening the micrograph file
            for i, particle in enumerate(selected_particles):
                idx = particle[0]
                # we stop and populate the image stack every time
                # we hit a particle whose location is this micrograph
                if idx == mrc_index:
                    cropped = self._crop_micrograph(arr, next(coord))
                    im[i] = cropped
        # Finally, apply transforms to resulting Image
        return self.generation_pipeline.forward(Image(im), indices)

    @staticmethod
    def _is_number(text):
        """
        Used in validation of coordinate files. We allow strings containing
        - or . to account for negative values and floats.
        """
        return text.replace("-", "1").replace(".", "1").isdigit()


class BoxesCoordinateSource(CoordinateSource):
    """
    Represents a data source consisting of micrographs and coordinate files
    in box format.
    """

    def __init__(
        self,
        files,
        particle_size=None,
        max_rows=None,
        B=0,
    ):
        """
        :param files: A list of tuples of the form (path_to_mrc, path_to_coord)
        :particle_size: Desired size of cropped particles (will override the size specified in coordinate file)
        :param max_rows: Maximum number of particles to read. (If `None`, will attempt to load all particles)
        """
        # instantiate super
        CoordinateSource.__init__(self, files, particle_size, max_rows, B)

    def _extract_box_size(self, box_file):
        with open(box_file, "r") as box:
            first_line = box.readlines()[0].split()
            if len(first_line) >= 4:
                box_size = int(float(first_line[2]))  # x size or y size works
                return box_size
            else:
                logger.error(f"Problem with coordinate file: {box_file}")
                raise ValueError(
                    "Coordinate file contains less than 4 numbers "
                    "per line. If these are particle centers, "
                    "use CentersCoordinateSource or  use the --centers "
                    "flag in aspire extract-particles."
                )

    def _validate_box_file(self, box_file, global_particle_size):
        with open(box_file, "r") as box:
            # validate each line, i.e. each particle
            for line in box.readlines():
                # box format requires 4 numbers per particle
                if len(line.split()) < 4:
                    logger.error(f"Problem with coordinate file: {box_file}")
                    raise ValueError(
                        "Coordinate file contains less than 4 numbers "
                        "per line. If these are particle centers, "
                        "use CentersCoordinateSource or  use the --centers "
                        "flag in aspire extract-particles."
                    )

                if not all(self._is_number(p) for p in line.split()):
                    logger.error(f"Problem with coordinate file: {box_file}")
                    raise ValueError(
                        "Coordinate file contains non-numeric coordinate values."
                    )

                # we can only accept square particles
                size_x, size_y = float(line.split()[2]), float(line.split()[3])
                if size_x != size_y:
                    logger.error(f"Problem with coordinate file: {box_file}")
                    raise ValueError(
                        "Coordinate file specifies non-square particle size "
                        f"{size_x}x{size_y}, but only square particles are supported."
                    )
                # check that this particle size is the *right* particle size
                if size_x != global_particle_size:
                    logger.error(f"Problem with coordinate file: {box_file}")
                    raise ValueError(
                        f"Coordinate file specifies a box size {size_x}x{size_x} "
                        "different from the box size found in the first "
                        f"coordinate file ({global_particle_size}x{global_particle_size}). "
                        "Particle size must be consistent."
                    )

    def _populate_particles(self, coord_paths):
        # overrides CoordinateSource._populate_particles because of the
        # possibility that force_new_particle_size will be called,
        # which requires self.particles to be populated already.
        # Also allows for validation of .box files prior to parsing them

        global_particle_size = self._extract_box_size(coord_paths[0])
        # validate the rest of the box files
        for coord_path in coord_paths:
            self._validate_box_file(coord_path, global_particle_size)

        # populate self.particles
        super()._populate_particles(coord_paths)

        # if particle size set by user, we have to re-do the coordinates
        if self.particle_size:
            self._force_new_particle_size(self.particle_size)

    def _coords_list_from_file(self, coord_file):
        """
        Given a coordinate file in box format, returns a list of coordinates.
        """
        with open(coord_file, "r") as infile:
            lines = [line.split() for line in infile.readlines()]
        # coords are already in box format, so simply cast to int
        return [[int(float(x)) for x in line] for line in lines]

    def _force_new_particle_size(self, new_size):
        """
        Given a new particle size, rewrite the coordinates so that the box size
        is changed, but still centered around the particle.
        """
        _resized_particles = []
        for particle in self.particles:
            mrc_index, box_coord = particle
            # get the coordinates of the center
            center = self._center_from_box_coord(box_coord)
            # rewrite to a box coordinate with new size
            new_coord = self._box_coord_from_center(center, new_size)
            _resized_particles.append((mrc_index, new_coord))
        self.particles = _resized_particles


class CentersCoordinateSource(CoordinateSource):
    """
    Represents a data source consisting of micrographs and coordinate files specifying particle centers only. Files can be text (.coord) or STAR files.
    """

    def __init__(self, files, particle_size, max_rows=None, B=0):
        """
        :param files: A list of tuples of the form (path_to_mrc, path_to_coord)
        :particle_size: Desired size of cropped particles (mandatory)
        :param max_rows: Maximum number of particles to read. (If `None`, will
        attempt to load all particles)
        """
        # instantiate super
        CoordinateSource.__init__(self, files, particle_size, max_rows, B)

    def _validate_centers_file(self, coord_file):
        """
        Ensures that a text file contains numeric particle centers.
        """
        with open(coord_file, "r") as infile:
            for line in infile.readlines():
                # need at least two numbers
                if len(line.split()) < 2:
                    logger.error(f"Problem with coordinate file: {coord_file}")
                    raise ValueError(
                        "Coordinate file contains a line with less than 2 numbers."
                    )
                # check that the coordinate has numeric values
                if not all(self._is_number(c) for c in line.split()):
                    logger.error(f"Problem with coordinate file: {coord_file}")
                    raise ValueError(
                        "Coordinate file contains non-numeric coordinate values."
                    )

    def _validate_starfile(self, coord_file):
        """
        Ensures that a STAR file contains numeric particle centers.
        """
        data_block = StarFile(coord_file).get_block_by_index(0)
        # We're looking for specific columns for the X and Y coordinates
        if not all(col in data_block for col in ["_rlnCoordinateX", "_rlnCoordinateY"]):
            logger.error(f"Problem with coordinate file: {coord_file}")
            raise ValueError(
                "STAR file does not contain _rlnCoordinateX, _rlnCoordinateY columns."
            )
        # check that all values in each column are numeric
        if not all(
            all(map(self._is_number, data_block[col]))
            for col in ["_rlnCoordinateX", "_rlnCoordinateY"]
        ):
            logger.error(f"Problem with coordinate file: {coord_file}")
            raise ValueError("STAR file contains non-numeric coordinate values.")

    def _populate_particles(self, coord_paths):
        # overrides CoordinateSource._populate_particles() in order
        # to validate coordinate files
        for coord_file in coord_paths:
            if os.path.splitext(coord_file)[1] == ".star":
                self._validate_starfile(coord_file)
            else:
                # assume text/.coord format
                self._validate_centers_file(coord_file)
        super()._populate_particles(coord_paths)

    def _coords_list_from_file(self, coord_file):
        """
        Given a coordinate file with (x,y) particle centers,
        return a list of coordinates in box format.
        """
        # check if it's a STAR file list of centers
        if os.path.splitext(coord_file)[1] == ".star":
            return self._coords_list_from_star(coord_file)
        # otherwise we assume text file format with one coord per line:
        with open(coord_file, "r") as infile:
            lines = [[float(c) for c in line.split()] for line in infile.readlines()]
        return [self._box_coord_from_center(line, self.particle_size) for line in lines]



import abc

import numpy as np
from numpy.linalg import qr

from aspire.utils import Rotation, bump_3d, grid_3d
from aspire.utils.random import Random, randn
from aspire.volume import Volume


class SyntheticVolumeBase(abc.ABC):
    def __init__(self, L, C, seed=None, dtype=np.float64):
        self.L = L
        self.C = C
        self.seed = seed
        self.dtype = dtype

    @abc.abstractmethod
    def generate(self):
        """
        Called to generate and return synthetic volumes.

        Each concrete subclass should implement this.
        """

    def __repr__(self):
        return f"{self.__class__.__name__} {self.__dict__}"


class GaussianBlobsVolume(SyntheticVolumeBase):
    """
    A base class for all volumes which are generated with randomized 3D Gaussians.
    """

    def __init__(self, L, C, K=16, alpha=1, seed=None, dtype=np.float64):
        """
        :param L: Resolution of the Volume(s) in pixels.
        :param C: Number of Volumes to generate.
        :param K: Number of Gaussian blobs used to construct the Volume(s).
        :param alpha: Scaling factor for variance of Gaussian blobs. Default=1.
        :param seed: Random seed for generating random Gaussian blobs.
        :param dtype: dtype for Volume(s)
        """
        self.K = int(K)
        self.alpha = float(alpha)
        super().__init__(L=L, C=C, seed=seed, dtype=dtype)

    @abc.abstractproperty
    def n_blobs(self):
        """
        The total number of Gaussian blobs used to generate a Volume.
        This value differs from `self.K` as it accounts for the blobs
        which have been duplicated during `_symmetrize_gaussians`.
        """

    @abc.abstractproperty
    def symmetry_group(self):
        """
        This property will be implemented by each subclass.
        """

    def generate(self):
        """
        Generates a Volume object with specified symmetry that is multiplied by a bump function
        to give compact support within the unit sphere.
        """
        vol = self._gaussian_blob_vols()

        bump_mask = bump_3d(self.L, spread=5, dtype=self.dtype)
        vol = np.multiply(bump_mask, vol)

        return Volume(vol)

    def _gaussian_blob_vols(self):
        """
        Generates a Volume object composed of Gaussian blobs.

        :return: A Volume instance containing C Gaussian blob volumes.
        """
        vols = np.zeros(shape=((self.C,) + (self.L,) * 3)).astype(self.dtype)
        with Random(self.seed):
            for c in range(self.C):
                Q, D, mu = self._gen_gaussians()
                Q_rot, D_sym, mu_rot = self._symmetrize_gaussians(Q, D, mu)
                vols[c] = self._eval_gaussians(Q_rot, D_sym, mu_rot)
        return Volume(vols)

    def _gen_gaussians(self):
        """
        For K gaussians, generate random orientation (Q), mean (mu), and variance (D).

        :return: Orientations Q, Variances D, Means mu.
        """
        Q = np.zeros(shape=(self.K, 3, 3)).astype(self.dtype)
        D = np.zeros(shape=(self.K, 3, 3)).astype(self.dtype)
        mu = np.zeros(shape=(self.K, 3)).astype(self.dtype)

        for k in range(self.K):
            V = randn(3, 3).astype(self.dtype) / np.sqrt(3)
            Q[k, :, :] = qr(V)[0]
            D[k, :, :] = (
                self.alpha**2 / self.n_blobs * np.diag(np.sum(abs(V) ** 2, axis=0))
            )
            mu[k, :] = 0.5 * randn(3) / np.sqrt(3)

        return Q, D, mu

    def _symmetrize_gaussians(self, Q, D, mu):
        """
        Called to add symmetry to Volumes by generating for each Gaussian blob duplicates in symmetric positions.
        """
        rots = self.symmetry_group.matrices

        Q_rot = np.zeros(shape=(self.n_blobs, 3, 3)).astype(self.dtype)
        D_sym = np.zeros(shape=(self.n_blobs, 3, 3)).astype(self.dtype)
        mu_rot = np.zeros(shape=(self.n_blobs, 3)).astype(self.dtype)
        idx = 0

        for rot in rots:
            for k in range(self.K):
                Q_rot[idx] = rot.T @ Q[k]
                D_sym[idx] = D[k]
                mu_rot[idx] = rot.T @ mu[k]
                idx += 1
        return Q_rot, D_sym, mu_rot

    def _eval_gaussians(self, Q, D, mu):
        """
        Evaluate Gaussian blobs over a 3D grid with centers, mu, orientations, Q, and variances, D.

        :param Q: A stack of size (n_blobs) x 3 x 3 of rotation matrices,
            determining the orientation of each blob.
        :param D: A stack of size (n_blobs) x 3 x 3 diagonal matrices,
            whose diagonal entries are the variances of each blob.
        :param mu: An array of size (n_blobs) x 3 containing the centers for each blob.

        :return: An L x L x L array.
        """
        g = grid_3d(self.L, indexing="xyz", dtype=self.dtype)
        coords = np.array(
            [g["x"].flatten(), g["y"].flatten(), g["z"].flatten()], dtype=self.dtype
        )

        n_blobs = Q.shape[0]
        vol = np.zeros(shape=(1, coords.shape[-1])).astype(self.dtype)

        for k in range(n_blobs):
            coords_k = coords - mu[k, :, np.newaxis]
            coords_k = (
                Q[k].T @ coords_k * np.sqrt(1 / np.diag(D[k, :, :]))[:, np.newaxis]
            )

            vol += np.exp(-0.5 * np.sum(np.abs(coords_k) ** 2, axis=0))

        vol = np.reshape(vol, g["x"].shape)

        return vol


class CnSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with cyclically symmetric volumes constructed of random 3D Gaussian blobs.
    """

    def __init__(self, L, C, order, K=16, alpha=1, seed=None, dtype=np.float64):
        """
        :param L: Resolution of the Volume(s) in pixels.
        :param C: Number of Volumes to generate.
        :param order: An integer representing the cyclic order of the Volume(s).
        :param K: Number of Gaussian blobs used to construct the Volume(s).
        :param seed: Random seed for generating random Gaussian blobs.
        :param dtype: dtype for Volume(s)
        """
        self.order = int(order)
        self._check_order()
        super().__init__(L=L, C=C, K=K, alpha=alpha, seed=seed, dtype=dtype)

    def _check_order(self):
        if self.order < 2:
            raise ValueError(
                f"For a {self.__class__.__name__} the cyclic order must be greater than 1. Provided order was {self.order}"
            )

    @property
    def n_blobs(self):
        return self.order * self.K

    @property
    def symmetry_group(self):
        """
        The Cn symmetry group contains all rotations about the z-axis
        by multiples of 2pi/n.

        In the case of an AsymmetricVolume or LegacyVolume, `symmetry_group`
        contains only the identity.

        :return: Rotation object containing the Cn symmetry group and the identity.
        """
        angles = np.zeros((self.order, 3), dtype=self.dtype)
        angles[:, 2] = 2 * np.pi * np.arange(self.order) / self.order
        return Rotation.from_euler(angles, dtype=self.dtype)


class DnSymmetricVolume(CnSymmetricVolume):
    """
    A Volume object with n-fold dihedral symmetry constructed of random 3D Gaussian blobs.
    """

    @property
    def n_blobs(self):
        return 2 * self.order * self.K

    @property
    def symmetry_group(self):
        """
        The Dn symmetry group contains all elements of the Cn symmetry group.
        In addition, for each element of the Cn symmetric group we rotate by
        pi about a perpendicular axis, in this case the y-axis.

        :return: Rotation object containing the Dn symmetry group and the identity.
        """
        # Rotations to induce cyclic symmetry
        angles = 2 * np.pi * np.arange(self.order, dtype=self.dtype) / self.order
        rot_z = Rotation.about_axis("z", angles).matrices

        # Perpendicular rotation to induce dihedral symmetry
        rot_perp = Rotation.about_axis("y", np.pi, dtype=self.dtype).matrices

        # Full set of rotations.
        rots = np.concatenate((rot_z, rot_z @ rot_perp[0].T), dtype=self.dtype)

        return Rotation(rots)


class TSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with tetrahedral symmetry constructed of random 3D Gaussian blobs.
    """

    @property
    def n_blobs(self):
        return 12 * self.K

    @property
    def symmetry_group(self):
        """
        A tetrahedron has C3 symmetry along the 4 axes through each vertex and
        perpendicular to the opposite face, and C2 symmetry along the axes through
        the midpoints of opposite edges. We convert from axis-angle representation of
        the symmetry group elements into rotation vectors to generate the rotation
        matrices via the `from_rotvec()` method.

        :return: Rotation object containing the tetrahedral symmetry group and the identity.
        """
        # C3 rotation vectors, ie. angle * axis.
        axes_C3 = np.array(
            [[1, 1, 1], [-1, -1, 1], [1, -1, -1], [-1, 1, -1]], dtype=self.dtype
        )
        axes_C3 /= np.linalg.norm(axes_C3, axis=-1)[..., np.newaxis]
        angles_C3 = np.array([2 * np.pi / 3, 4 * np.pi / 3], dtype=self.dtype)
        rot_vecs_C3 = np.array(
            [angle * axes_C3 for angle in angles_C3], dtype=self.dtype
        ).reshape((8, 3))

        # C2 rotation vectors.
        axes_C2 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=self.dtype)
        rot_vecs_C2 = np.pi * axes_C2

        # The full set of rotation vectors inducing tetrahedral symmetry.
        rot_vec_I = np.zeros((1, 3), dtype=self.dtype)
        rot_vecs = np.concatenate(
            (rot_vec_I, rot_vecs_C3, rot_vecs_C2), dtype=self.dtype
        )

        # Return rotations.
        return Rotation.from_rotvec(rot_vecs, dtype=self.dtype)


class OSymmetricVolume(GaussianBlobsVolume):
    """
    A Volume object with octahedral symmetry constructed of random 3D Gaussian blobs.
    """

    @property
    def n_blobs(self):
        return 24 * self.K

    @property
    def symmetry_group(self):
        """
        The symmetry group elements of the octahedral symmetry group O are the identity,
        the elements of 3 C4 rotation groups whose axes pass through two opposite vertices of the
        regular octahedron, 4 C3 rotation groups whose axes pass through the midpoints of two of
        its opposite faces, and 6 C2 rotation groups whose axes pass through the midpoints of two of
        its opposite edges.

        :return: Rotation object containing the octahedral symmetry group and the identity.
        """

        # C4 rotation vectors, ie angle * axis
        axes_C4 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=self.dtype)
        angles_C4 = np.array([np.pi / 2, np.pi, 3 * np.pi / 2], dtype=self.dtype)
        rot_vecs_C4 = np.array(
            [angle * axes_C4 for angle in angles_C4], dtype=self.dtype
        ).reshape((9, 3))

        # C3 rotation vectors
        axes_C3 = np.array(
            [[1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1]], dtype=self.dtype
        )
        axes_C3 /= np.linalg.norm(axes_C3, axis=-1)[..., np.newaxis]
        angles_C3 = np.array([2 * np.pi / 3, 4 * np.pi / 3], dtype=self.dtype)
        rot_vecs_C3 = np.array(
            [angle * axes_C3 for angle in angles_C3], dtype=self.dtype
        ).reshape((8, 3))

        # C2 rotation vectors
        axes_C2 = np.array(
            [[1, 1, 0], [-1, 1, 0], [1, 0, 1], [-1, 0, 1], [0, 1, 1], [0, -1, 1]],
            dtype=self.dtype,
        )
        axes_C2 /= np.linalg.norm(axes_C2, axis=-1)[..., np.newaxis]
        rot_vecs_C2 = np.pi * axes_C2

        # The full set of rotation vectors inducing octahedral symmetry.
        rot_vec_I = np.zeros((1, 3), dtype=self.dtype)
        rot_vecs = np.concatenate(
            (rot_vec_I, rot_vecs_C4, rot_vecs_C3, rot_vecs_C2), dtype=self.dtype
        )

        # Return rotations.
        return Rotation.from_rotvec(rot_vecs, dtype=self.dtype)


class AsymmetricVolume(CnSymmetricVolume):
    """
    An asymmetric Volume constructed of random 3D Gaussian blobs with compact support in the unit sphere.
    """

    def __init__(self, L, C, K=64, seed=None, dtype=np.float64):
        super().__init__(L=L, C=C, K=K, order=1, seed=seed, dtype=dtype)

    def _check_order(self):
        if self.order != 1:
            raise ValueError(
                f"An {self.__class__.__name__} must have order=1. Provided order was {self.order}"
            )

    def _symmetrize_gaussians(self, Q, D, mu):
        return Q, D, mu


class LegacyVolume(AsymmetricVolume):
    """
    An asymmetric Volume object used for testing of legacy code.
    """

    def __init__(self, L, C=2, K=16, seed=0, dtype=np.float64):
        super().__init__(L=L, C=C, K=K, seed=seed, dtype=dtype)

    def generate(self):
        """
        Generates an asymmetric volume composed of random 3D Gaussian blobs.
        """
        return self._gaussian_blob_vols()



import logging

from aspire.abinitio import CLOrient3D

logger = logging.getLogger(__name__)


class CommLineSDP(CLOrient3D):
    """
    Class to estimate 3D orientations using Semi-Definite Programming
    :cite:`DBLP:journals/siamis/SingerS11`
    """

    def __init__(self, src):
        """
        constructor of an object for estimating 3D orientations
        """
        pass

    def estimate(self):
        """
        perform estimation of orientations
        """
        pass

    def output(self):
        """
        Output the 3D orientations
        """
        pass



import logging

import numpy as np
from numpy.linalg import eig, inv
from scipy.linalg import solve, sqrtm

from aspire.operators import BlkDiagMatrix, RadialCTFFilter
from aspire.optimization import conj_grad, fill_struct
from aspire.utils import make_symmat
from aspire.utils.matlab_compat import m_reshape

logger = logging.getLogger(__name__)


def shrink_covar(covar, noise_var, gamma, shrinker="frobenius_norm"):
    """
    Shrink the covariance matrix

    :param covar_in: An input covariance matrix
    :param noise_var: The estimated variance of noise
    :param gamma: An input parameter to specify the maximum values of eigen values to be neglected.
    :param shrinker: An input parameter to select different shrinking methods.
    :return: The shrinked covariance matrix
    """

    assert shrinker in (
        "frobenius_norm",
        "operator_norm",
        "soft_threshold",
    ), "Unsupported shrink method"

    lambs, eig_vec = eig(make_symmat(covar))

    lambda_max = noise_var * (1 + np.sqrt(gamma)) ** 2

    lambs[lambs < lambda_max] = 0

    if shrinker == "operator_norm":
        lambdas = lambs[lambs > lambda_max]
        lambdas = (
            1
            / 2
            * (
                lambdas
                - noise_var * (gamma - 1)
                + np.sqrt(
                    (lambdas - noise_var * (gamma - 1)) ** 2 - 4 * noise_var * lambdas
                )
            )
            - noise_var
        )

        lambs[lambs > lambda_max] = lambdas
    elif shrinker == "frobenius_norm":
        lambdas = lambs[lambs > lambda_max]
        lambdas = (
            1
            / 2
            * (
                lambdas
                - noise_var * (gamma - 1)
                + np.sqrt(
                    (lambdas - noise_var * (gamma - 1)) ** 2 - 4 * noise_var * lambdas
                )
            )
            - noise_var
        )
        c = np.divide(
            (1 - np.divide(noise_var**2 * gamma, lambdas**2)),
            (1 + np.divide(noise_var * gamma, lambdas)),
        )
        lambdas = lambdas * c
        lambs[lambs > lambda_max] = lambdas
    else:
        # for the case of shrinker == 'soft_threshold'
        lambdas = lambs[lambs > lambda_max]
        lambs[lambs > lambda_max] = lambdas - lambda_max

    diag_lambs = np.zeros_like(covar)
    np.fill_diagonal(diag_lambs, lambs)

    shrinked_covar = eig_vec @ diag_lambs @ eig_vec.conj().T

    return shrinked_covar


class RotCov2D:
    """
    Define a class for performing Cov2D analysis with CTF information described in

    T. Bhamre, T. Zhang, and A. Singer, "Denoising and covariance estimation of single particle cryo-EM images",
    J. Struct. Biol. 195, 27-81 (2016). DOI: 10.1016/j.jsb.2016.04.013
    """

    def __init__(self, basis):
        """
        constructor of an object for 2D covariance analysis
        """
        self.basis = basis
        self.dtype = self.basis.dtype
        assert basis.ndim == 2, "Only two-dimensional basis functions are needed."

    def _get_mean(self, coeffs):
        """
        Calculate the mean vector from the expansion coefficients of 2D images without CTF information.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be averaged.
        :return: The mean value vector for all images.
        """
        if coeffs.size == 0:
            raise RuntimeError("The coefficients need to be calculated first!")
        mask = self.basis._indices["ells"] == 0
        mean_coeff = np.zeros(self.basis.count, dtype=coeffs.dtype)
        mean_coeff[mask] = np.mean(coeffs[..., mask], axis=0)

        return mean_coeff

    def _get_covar(self, coeffs, mean_coeff=None, do_refl=True):
        """
        Calculate the covariance matrix from the expansion coefficients without CTF information.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) calculated from 2D images.
        :param mean_coeff: The mean vector calculated from the `coeffs`.
        :param do_refl: If true, enforce invariance to reflection (default false).
        :return: The covariance matrix of coefficients for all images.
        """
        if coeffs.size == 0:
            raise RuntimeError("The coefficients need to be calculated first!")
        if mean_coeff is None:
            mean_coeff = self._get_mean(coeffs)

        # Initialize a totally empty BlkDiagMatrix, build incrementally.
        covar_coeff = BlkDiagMatrix.empty(0, dtype=coeffs.dtype)
        ell = 0
        mask = self.basis._indices["ells"] == ell
        coeff_ell = coeffs[..., mask] - mean_coeff[mask]
        covar_ell = np.array(coeff_ell.T @ coeff_ell / coeffs.shape[0])
        covar_coeff.append(covar_ell)

        for ell in range(1, self.basis.ell_max + 1):
            mask_ell = self.basis._indices["ells"] == ell
            mask_pos = mask_ell & (self.basis._indices["sgns"] == +1)
            mask_neg = mask_ell & (self.basis._indices["sgns"] == -1)

            covar_ell_diag = np.array(
                coeffs[:, mask_pos].T @ coeffs[:, mask_pos]
                + coeffs[:, mask_neg].T @ coeffs[:, mask_neg]
            ) / (2 * coeffs.shape[0])

            if do_refl:
                covar_coeff.append(covar_ell_diag)
                covar_coeff.append(covar_ell_diag)
            else:
                covar_ell_off = np.array(
                    (
                        coeffs[:, mask_pos] @ coeffs[:, mask_neg].T / coeffs.shape[0]
                        - coeffs[:, mask_pos].T @ coeffs[:, mask_neg]
                    )
                    / (2 * coeffs.shape[0])
                )

                hsize = covar_ell_diag.shape[0]
                covar_coeff_blk = np.zeros((2, hsize, 2, hsize))

                covar_coeff_blk[0:2, :, 0:2, :] = covar_ell_diag[:hsize, :hsize]
                covar_coeff_blk[0, :, 1, :] = covar_ell_off[:hsize, :hsize]
                covar_coeff_blk[1, :, 0, :] = covar_ell_off.T[:hsize, :hsize]

                covar_coeff.append(covar_coeff_blk.reshape(2 * hsize, 2 * hsize))

        return covar_coeff

    def get_mean(self, coeffs, ctf_fb=None, ctf_idx=None):
        """
        Calculate the mean vector from the expansion coefficients with CTF information.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be averaged.
        :param ctf_fb: The CFT functions in the FB expansion.
        :param ctf_idx: An array of the CFT function indices for all 2D images.
            If ctf_fb or ctf_idx is None, the identity filter will be applied.
        :return: The mean value vector for all images.
        """

        if coeffs.size == 0:
            raise RuntimeError("The coefficients need to be calculated!")

        # should assert we require none or both...
        if (ctf_fb is None) or (ctf_idx is None):
            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
            ctf_fb = [
                BlkDiagMatrix.eye_like(
                    RadialCTFFilter().fb_mat(self.basis), dtype=self.dtype
                )
            ]

        b = np.zeros(self.basis.count, dtype=coeffs.dtype)

        A = BlkDiagMatrix.zeros_like(ctf_fb[0])
        for k in np.unique(ctf_idx[:]).T:
            coeff_k = coeffs[ctf_idx == k]
            weight = coeff_k.shape[0] / coeffs.shape[0]
            mean_coeff_k = self._get_mean(coeff_k)

            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T
            b += weight * ctf_fb_k_t.apply(mean_coeff_k)
            A += weight * (ctf_fb_k_t @ ctf_fb_k)

        mean_coeff = A.solve(b)
        return mean_coeff

    def get_covar(
        self,
        coeffs,
        ctf_fb=None,
        ctf_idx=None,
        mean_coeff=None,
        do_refl=True,
        noise_var=0,
        covar_est_opt=None,
        make_psd=True,
    ):
        """
        Calculate the covariance matrix from the expansion coefficients and CTF information.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
        :param ctf_fb: The CFT functions in the FB expansion.
        :param ctf_idx: An array of the CFT function indices for all 2D images.
            If ctf_fb or ctf_idx is None, the identity filter will be applied.
        :param mean_coeff: The mean value vector from all images.
        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
            from clean images of simulation data.
        :param covar_est_opt: The optimization parameter list for obtaining the Cov2D matrix.
        :param make_psd: If True, make the covariance matrix positive semidefinite
        :return: The basis coefficients of the covariance matrix in
            the form of cell array representing a block diagonal matrix. These
            block diagonal matrices are implemented as BlkDiagMatrix instances.
            The covariance is calculated from the images represented by the coeffs array,
            along with all possible rotations and reflections. As a result, the computed covariance
            matrix is invariant to both reflection and rotation. The effect of the filters in ctf_fb
            are accounted for and inverted to yield a covariance estimate of the unfiltered images.
        """

        if coeffs.size == 0:
            raise RuntimeError("The coefficients need to be calculated!")

        if (ctf_fb is None) or (ctf_idx is None):
            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
            ctf_fb = [
                BlkDiagMatrix.eye_like(
                    RadialCTFFilter().fb_mat(self.basis), dtype=self.dtype
                )
            ]

        def identity(x):
            return x

        default_est_opt = {
            "shrinker": None,
            "verbose": 0,
            "max_iter": 250,
            "iter_callback": [],
            "store_iterates": False,
            "rel_tolerance": 1e-12,
            "precision": self.dtype,
            "preconditioner": identity,
        }

        covar_est_opt = fill_struct(covar_est_opt, default_est_opt)

        if mean_coeff is None:
            mean_coeff = self.get_mean(coeffs, ctf_fb, ctf_idx)

        b_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])
        b_noise = BlkDiagMatrix.zeros_like(ctf_fb[0])
        A = []
        for _ in range(len(ctf_fb)):
            A.append(BlkDiagMatrix.zeros_like(ctf_fb[0]))

        M = BlkDiagMatrix.zeros_like(ctf_fb[0])

        for k in np.unique(ctf_idx[:]):
            coeff_k = coeffs[ctf_idx == k].astype(self.dtype)
            weight = coeff_k.shape[0] / coeffs.shape[0]

            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T
            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
            covar_coeff_k = self._get_covar(coeff_k, mean_coeff_k)

            b_coeff += weight * (ctf_fb_k_t @ covar_coeff_k @ ctf_fb_k)

            ctf_fb_k_sq = ctf_fb_k_t @ ctf_fb_k
            b_noise += weight * ctf_fb_k_sq

            A[k] = np.sqrt(weight) * ctf_fb_k_sq
            M += A[k]

        if not b_coeff.check_psd():
            logger.warning("Left side b in Cov2D is not positive semidefinite.")

        if covar_est_opt["shrinker"] is None:
            b = b_coeff - noise_var * b_noise
        else:
            b = self.shrink_covar_backward(
                b_coeff,
                b_noise,
                np.size(coeffs, 0),
                noise_var,
                covar_est_opt["shrinker"],
            )
        if not b.check_psd():
            logger.warning(
                "Left side b after removing noise in Cov2D"
                " is not positive semidefinite."
            )

        # RCOPT okay, this looks like a big batch, come back later

        cg_opt = covar_est_opt

        covar_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])

        def precond_fun(S, x):
            p = np.size(S, 0)
            assert np.size(x) == p * p, "The sizes of S and x are not consistent."
            x = m_reshape(x, (p, p))
            y = S @ x @ S
            y = m_reshape(y, (p**2,))
            return y

        def apply(A, x):
            p = np.size(A[0], 0)
            x = m_reshape(x, (p, p))
            y = np.zeros_like(x)
            for k in range(0, len(A)):
                y = y + A[k] @ x @ A[k].T
            y = m_reshape(y, (p**2,))
            return y

        for ell in range(0, len(b)):
            A_ell = []
            for k in range(0, len(A)):
                A_ell.append(A[k][ell])
            p = np.size(A_ell[0], 0)
            b_ell = m_reshape(b[ell], (p**2,))
            S = inv(M[ell])
            cg_opt["preconditioner"] = lambda x, S=S: precond_fun(S, x)
            covar_coeff_ell, _, _ = conj_grad(
                lambda x, A_ell=A_ell: apply(A_ell, x), b_ell, cg_opt
            )
            covar_coeff[ell] = m_reshape(covar_coeff_ell, (p, p))

        if not covar_coeff.check_psd():
            logger.warning("Covariance matrix in Cov2D is not positive semidefinite.")
            if make_psd:
                logger.info("Convert matrices to positive semidefinite.")
                covar_coeff = covar_coeff.make_psd()

        return covar_coeff

    def shrink_covar_backward(self, b, b_noise, n, noise_var, shrinker):
        """
        Apply the shrinking method to the 2D covariance of coefficients.

        :param b: An input coefficient covariance.
        :param b_noise: The noise covariance.
        :param noise_var: The estimated variance of noise.
        :param shrinker: The shrinking method.
        :return: The shrinked 2D covariance coefficients.
        """
        b_out = b
        for ell in range(0, len(b)):
            b_ell = b[ell]
            p = np.size(b_ell, 1)
            # scipy >= 1.6.0 will upcast the sqrtm result to doubles
            #  https://github.com/scipy/scipy/issues/14853
            S = sqrtm(b_noise[ell]).astype(self.dtype)
            # from Matlab b_ell = S \ b_ell /S
            b_ell = solve(S, b_ell) @ inv(S)
            b_ell = shrink_covar(b_ell, noise_var, p / n, shrinker)
            b_ell = S @ b_ell @ S
            b_out[ell] = b_ell
        return b_out

    def get_cwf_coeffs(
        self,
        coeffs,
        ctf_fb=None,
        ctf_idx=None,
        mean_coeff=None,
        covar_coeff=None,
        noise_var=0,
    ):
        """
        Estimate the expansion coefficients using the Covariance Wiener Filtering (CWF) method.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
        :param ctf_fb: The CFT functions in the FB expansion.
        :param ctf_idx: An array of the CFT function indices for all 2D images.
            If ctf_fb or ctf_idx is None, the identity filter will be applied.
        :param mean_coeff: The mean value vector from all images.
        :param covar_coeff: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
            from clean images of simulation data.
        :return: The estimated coefficients of the unfiltered images in certain math basis.
            These are obtained using a Wiener filter with the specified covariance for the clean images
            and white noise of variance `noise_var` for the noise.
        """

        if mean_coeff is None:
            mean_coeff = self.get_mean(coeffs, ctf_fb, ctf_idx)

        if covar_coeff is None:
            covar_coeff = self.get_covar(
                coeffs, ctf_fb, ctf_idx, mean_coeff, noise_var=noise_var
            )

        # Handle CTF arguments.
        if (ctf_fb is None) ^ (ctf_idx is None):
            raise RuntimeError(
                "Both `ctf_fb` and `ctf_idx` should be provided,"
                " or both should be `None`."
                f' Given {"ctf_fb" if ctf_idx is None else "ctf_idx"}'
            )
        elif ctf_fb is None:
            # Setup defaults for CTF
            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
            ctf_fb = [BlkDiagMatrix.eye_like(covar_coeff)]

        noise_covar_coeff = noise_var * BlkDiagMatrix.eye_like(covar_coeff)

        coeffs_est = np.zeros_like(coeffs)

        for k in np.unique(ctf_idx[:]):
            coeff_k = coeffs[ctf_idx == k]
            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T

            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
            coeff_est_k = coeff_k - mean_coeff_k

            if noise_var == 0:
                coeff_est_k = ctf_fb_k.solve(coeff_est_k.T).T
            else:
                sig_covar_coeff = ctf_fb_k @ covar_coeff @ ctf_fb_k_t
                sig_noise_covar_coeff = sig_covar_coeff + noise_covar_coeff

                coeff_est_k = sig_noise_covar_coeff.solve(coeff_est_k.T).T
                coeff_est_k = (covar_coeff @ ctf_fb_k_t).apply(coeff_est_k.T).T

            coeff_est_k = coeff_est_k + mean_coeff
            coeffs_est[ctf_idx == k] = coeff_est_k

        return coeffs_est


class BatchedRotCov2D(RotCov2D):
    """
    Perform batchwise rotationally equivariant 2D covariance estimation from an
    `ImageSource` objects. This is done with a single pass through the data,
    processing moderately-sized batches one at a time. The rotational
    equivariance is achieved by decomposing images in a steerable Fourier–Bessel
    basis. For more information, see

        T. Bhamre, T. Zhang, and A. Singer, "Denoising and covariance estimation
        of single particle cryo-EM images", J. Struct. Biol. 195, 27-81 (2016).
        DOI: 10.1016/j.jsb.2016.04.013

    :param src: The `ImageSource` object from which the sample images are to
        be extracted.
    :param basis: The `FBBasis2D` object used to decompose the images. By
        default, this is set to `FFBBasis2D((src.L, src.L))`.
    :param batch_size: The number of images to process at a time (default 8192).
    """

    def __init__(self, src, basis=None, batch_size=8192):
        self.src = src
        self.basis = basis
        self.batch_size = batch_size
        self.dtype = self.src.dtype

        self.b_mean = None
        self.b_covar = None
        self.A_mean = None
        self.A_covar = None
        self.M_covar = None

        self._build()

    def _build(self):
        src = self.src

        if self.basis is None:
            from aspire.basis import FFBBasis2D

            self.basis = FFBBasis2D((src.L, src.L), dtype=self.dtype)

        if not src.unique_filters:
            logger.info("CTF filters are not included in Cov2D denoising")
            # set all CTF filters to an identity filter
            self.ctf_idx = np.zeros(src.n, dtype=int)
            self.ctf_fb = [BlkDiagMatrix.eye_like(RadialCTFFilter().fb_mat(self.basis))]
        else:
            logger.info("Represent CTF filters in FB basis")
            unique_filters = src.unique_filters
            self.ctf_idx = src.filter_indices
            self.ctf_fb = [f.fb_mat(self.basis) for f in unique_filters]

    def _calc_rhs(self):
        src = self.src
        basis = self.basis

        ctf_fb = self.ctf_fb
        ctf_idx = self.ctf_idx

        zero_coeff = np.zeros((basis.count,), dtype=self.dtype)

        b_mean = [np.zeros(basis.count, dtype=self.dtype) for _ in ctf_fb]

        b_covar = BlkDiagMatrix.zeros_like(ctf_fb[0])

        for start in range(0, src.n, self.batch_size):
            batch = np.arange(start, min(start + self.batch_size, src.n))

            im = src.images[batch[0] : batch[0] + len(batch)]
            coeff = basis.evaluate_t(im)

            for k in np.unique(ctf_idx[batch]):
                coeff_k = coeff[ctf_idx[batch] == k]
                weight = np.size(coeff_k, 0) / src.n

                mean_coeff_k = self._get_mean(coeff_k)

                ctf_fb_k = ctf_fb[k]
                ctf_fb_k_t = ctf_fb_k.T

                b_mean_k = weight * ctf_fb_k_t.apply(mean_coeff_k)

                b_mean[k] += b_mean_k

                covar_coeff_k = self._get_covar(coeff_k, zero_coeff)

                b_covar_k = ctf_fb_k_t @ covar_coeff_k
                b_covar_k = b_covar_k @ ctf_fb_k
                b_covar_k *= weight

                b_covar += b_covar_k

        self.b_mean = b_mean
        self.b_covar = b_covar

    def _calc_op(self):
        src = self.src

        ctf_fb = self.ctf_fb
        ctf_idx = self.ctf_idx

        A_mean = BlkDiagMatrix.zeros_like(ctf_fb[0])
        A_covar = [None for _ in ctf_fb]
        M_covar = BlkDiagMatrix.zeros_like(ctf_fb[0])

        for k in np.unique(ctf_idx):
            weight = np.count_nonzero(ctf_idx == k) / src.n

            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T

            ctf_fb_k_sq = ctf_fb_k_t @ ctf_fb_k
            A_mean_k = weight * ctf_fb_k_sq
            A_mean += A_mean_k

            A_covar_k = np.sqrt(weight) * ctf_fb_k_sq
            A_covar[k] = A_covar_k

            M_covar += A_covar_k

        self.A_mean = A_mean
        self.A_covar = A_covar
        self.M_covar = M_covar

    def _mean_correct_covar_rhs(self, b_covar, b_mean, mean_coeff):
        src = self.src

        ctf_fb = self.ctf_fb
        ctf_idx = self.ctf_idx

        partition = ctf_fb[0].partition

        # Note: If we don't do this, we'll be modifying the stored `b_covar`
        # since the operations below are in-place.
        b_covar = b_covar.copy()

        for k in np.unique(ctf_idx):
            weight = np.count_nonzero(ctf_idx == k) / src.n

            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T

            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
            mean_coeff_k = ctf_fb_k_t.apply(mean_coeff_k)

            mean_coeff_k = mean_coeff_k[: partition[0][0]]
            b_mean_k = b_mean[k][: partition[0][0]]

            correction = (
                np.outer(mean_coeff_k, b_mean_k)
                + np.outer(b_mean_k, mean_coeff_k)
                - weight * np.outer(mean_coeff_k, mean_coeff_k)
            )

            b_covar[0] -= correction

        return b_covar

    def _noise_correct_covar_rhs(self, b_covar, b_noise, noise_var, shrinker):
        if shrinker is None:
            b_noise = -noise_var * b_noise
            b_covar += b_noise
        else:
            b_covar = self.shrink_covar_backward(
                b_covar, b_noise, self.src.n, noise_var, shrinker
            )

        return b_covar

    def _solve_covar(self, A_covar, b_covar, M, covar_est_opt):
        ctf_fb = self.ctf_fb

        def precond_fun(S, x):
            p = np.size(S, 0)
            assert np.size(x) == p * p, "The sizes of S and x are not consistent."
            x = m_reshape(x, (p, p))
            y = S @ x @ S
            y = m_reshape(y, (p**2,))
            return y

        def apply(A, x):
            p = np.size(A[0], 0)
            x = m_reshape(x, (p, p))
            y = np.zeros_like(x)
            for k in range(0, len(A)):
                y = y + A[k] @ x @ A[k].T
            y = m_reshape(y, (p**2,))
            return y

        cg_opt = covar_est_opt
        covar_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])

        for ell in range(0, len(b_covar)):
            A_ell = []
            for k in range(0, len(A_covar)):
                A_ell.append(A_covar[k][ell])
            p = np.size(A_ell[0], 0)
            b_ell = m_reshape(b_covar[ell], (p**2,))
            S = inv(M[ell])
            cg_opt["preconditioner"] = lambda x, S=S: precond_fun(S, x)
            covar_coeff_ell, _, _ = conj_grad(
                lambda x, A_ell=A_ell: apply(A_ell, x), b_ell, cg_opt
            )
            covar_coeff[ell] = m_reshape(covar_coeff_ell, (p, p))

        return covar_coeff

    def get_mean(self):
        """
        Calculate the rotationally invariant mean image in the basis
        coefficients.

        :return: The mean coefficient vector in `self.basis`.
        """

        if not self.b_mean:
            self._calc_rhs()

        if not self.A_mean:
            self._calc_op()

        b_mean_all = np.stack(self.b_mean).sum(axis=0)
        mean_coeff = self.A_mean.solve(b_mean_all)

        return mean_coeff

    def get_covar(
        self, noise_var=0, mean_coeff=None, covar_est_opt=None, make_psd=True
    ):
        """
        Calculate the block diagonal covariance matrix in the basis
        coefficients.

        :param noise_var: The variance of the noise in the images (default 1)
        :param mean_coeff: If specified, overrides the mean coefficient vector
            used to calculate the covariance (default `self.get_mean()`).
        :param :covar_est_opt: The estimation parameters for obtaining the covariance
            matrix in the form of a dictionary. Keys include:
            - 'shrinker': The type of shrinkage we apply to the right-hand side
              in the normal equations. Can be `'None'`, in which case no
              shrinkage is performed. For a list of shrinkers, see the
              documentation of `shrink_covar`.
            - 'verbose': Verbosity (integer) of the conjugate gradient algorithm
              (see documentation for `conj_grad`, default zero).
            - 'max_iter': Maximum number of conjugate gradient iterations (see
              documentation for `conj_grad`, default 250).
            - 'iter_callback': Callback performed at the end of an iteration
              (see documentation for `conj_grad`, default `[]`).
            - 'store_iterates': Determines whether to store intermediate
              iterates (see documentation for `conj_grad`, default `False`).
            - 'rel_tolerance': Relative stopping tolerance of the conjugate
              gradient algorithm (see documentation for `conj_grad`, default
              `1e-12`).
            - 'precision': Precision of conjugate gradient algorithm (see
              documentation for `conj_grad`, default `'float64'`)
        :param make_psd: If True, make the covariance matrix positive semidefinite
        :return: The block diagonal matrix containing the basis coefficients (in
            `self.basis`) for the estimated covariance matrix. These are
            implemented using `BlkDiagMatrix`.
        """

        def identity(x):
            return x

        default_est_opt = {
            "shrinker": None,
            "verbose": 0,
            "max_iter": 250,
            "iter_callback": [],
            "store_iterates": False,
            "rel_tolerance": 1e-12,
            "precision": self.dtype,
            "preconditioner": identity,
        }

        covar_est_opt = fill_struct(covar_est_opt, default_est_opt)

        if not self.b_covar:
            self._calc_rhs()

        if not self.A_covar or self.M_covar:
            self._calc_op()

        if mean_coeff is None:
            mean_coeff = self.get_mean()

        b_covar = self.b_covar

        b_covar = self._mean_correct_covar_rhs(b_covar, self.b_mean, mean_coeff)
        if not b_covar.check_psd():
            logger.warning("Left side b in Batched Cov2D is not positive semidefinite.")

        b_covar = self._noise_correct_covar_rhs(
            b_covar, self.A_mean, noise_var, covar_est_opt["shrinker"]
        )
        if not b_covar.check_psd():
            logger.warning(
                "Left side b after removing noise "
                "in Batched Cov2D is not positive semidefinite."
            )

        covar_coeff = self._solve_covar(
            self.A_covar, b_covar, self.M_covar, covar_est_opt
        )
        if not covar_coeff.check_psd():
            logger.warning(
                "Covariance matrix in Batched Cov2D is not positive semidefinite."
            )
            if make_psd:
                logger.info("Convert matrices to positive semidefinite.")
                covar_coeff = covar_coeff.make_psd()

        return covar_coeff

    def get_cwf_coeffs(
        self, coeffs, ctf_fb, ctf_idx, mean_coeff, covar_coeff, noise_var=0
    ):
        """
        Estimate the expansion coefficients using the Covariance Wiener Filtering (CWF) method.

        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
        :param ctf_fb: The CFT functions in the FB expansion.
        :param ctf_idx: An array of the CFT function indices for all 2D images.
            If ctf_fb or ctf_idx is None, the identity filter will be applied.
        :param mean_coeff: The mean value vector from all images.
        :param covar_coeff: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
            from clean images of simulation data.
        :return: The estimated coefficients of the unfiltered images in certain math basis.
            These are obtained using a Wiener filter with the specified covariance for the clean images
            and white noise of variance `noise_var` for the noise.
        """

        if mean_coeff is None:
            mean_coeff = self.get_mean()

        if covar_coeff is None:
            covar_coeff = self.get_covar(noise_var=noise_var, mean_coeff=mean_coeff)

        # Handle CTF arguments.
        if (ctf_fb is None) ^ (ctf_idx is None):
            raise RuntimeError(
                "Both `ctf_fb` and `ctf_idx` should be provided,"
                " or both should be `None`."
                f' Given {"ctf_fb" if ctf_idx is None else "ctf_idx"}'
            )
        elif ctf_fb is None:
            # Setup defaults for CTF
            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
            ctf_fb = [BlkDiagMatrix.eye_like(covar_coeff)]

        noise_covar_coeff = noise_var * BlkDiagMatrix.eye_like(covar_coeff)

        coeffs_est = np.zeros_like(coeffs)

        for k in np.unique(ctf_idx[:]):
            coeff_k = coeffs[ctf_idx == k]
            ctf_fb_k = ctf_fb[k]
            ctf_fb_k_t = ctf_fb_k.T

            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
            coeff_est_k = coeff_k - mean_coeff_k

            if noise_var == 0:
                coeff_est_k = ctf_fb_k.solve(coeff_est_k.T).T
            else:
                sig_covar_coeff = ctf_fb_k @ covar_coeff @ ctf_fb_k_t
                sig_noise_covar_coeff = sig_covar_coeff + noise_covar_coeff

                coeff_est_k = sig_noise_covar_coeff.solve(coeff_est_k.T).T
                coeff_est_k = (covar_coeff @ ctf_fb_k_t).apply(coeff_est_k.T).T

            coeff_est_k = coeff_est_k + mean_coeff
            coeffs_est[ctf_idx == k] = coeff_est_k

        return coeffs_est



import logging

import numpy as np
from skimage.filters import difference_of_gaussians, window
from skimage.transform import rotate, warp_polar

from aspire.numeric import fft
from aspire.utils.coor_trans import grid_2d

logger = logging.getLogger(__name__)

__cache = dict()


def _phase_cross_correlation(img0, img1):
    """
    # Adapted from skimage.registration.phase_cross_correlation

    :param img0: Fixed image.
    :param img1: Translated image.
    :returns: (cross-correlation magnitudes (2D array), shifts)
    """

    # Cache img0 transform, this saves n_classes*(n_nbor-1) transforms
    # Note we use the `id` because ndarray are unhashable
    key = id(img0)
    if key not in __cache:
        __cache[key] = fft.fft2(img0)
    src_f = __cache[key]

    target_f = fft.fft2(img1)

    # Whole-pixel shifts - Compute cross-correlation by an IFFT
    shape = src_f.shape
    image_product = src_f * target_f.conj()
    cross_correlation = fft.ifft2(image_product)

    # Locate maximum
    maxima = np.unravel_index(
        np.argmax(np.abs(cross_correlation)), cross_correlation.shape
    )
    midpoints = np.array([np.fix(axis_size / 2) for axis_size in shape])

    shifts = np.array(maxima, dtype=np.float64)
    shifts[shifts > midpoints] -= np.array(shape)[shifts > midpoints]

    return np.abs(cross_correlation), shifts


def reddy_chatterji_register(
    images, reflection, mask=None, do_cross_corr_translations=True, dtype=None
):
    """
    Compute the Reddy Chatterji method registering images[1:] to images[0].

    This differs from papers and published scikit implimentations by
    computing the fixed base image[0] pipeline once then reusing.

    :param images: Image data (m_img, L, L)
    :param reflection: Image reflections (m_img,)
    :param mask: Support of image. Defaults to disk with radius images.shape[-1]//2.
    :param do_cross_corr_translations: Solve trnaslations by using cross correlation (log polar) method.
    :param dtype: Specify dtype.  Defaults to infer from images.dtype
    :returns: (rotations, shifts, correlations) corresponding to `images`
    """

    if mask is None:
        L = images.shape[-1]
        mask = grid_2d(L, normalized=False)["r"] < L // 2
    if dtype is None:
        dtype = images.dtype

    # Result arrays
    M = len(images)
    rotations = np.zeros(M, dtype=dtype)
    correlations = np.full(M, -np.inf, dtype=dtype)
    shifts = np.zeros((M, 2), dtype=int)

    # De-Mean
    images = images - images.mean(axis=(-1, -2))[:, np.newaxis, np.newaxis]

    # Precompute fixed_img data used repeatedly in the loop below.
    fixed_img = images[0]
    # Difference of Gaussians (Band Filter)
    fixed_img_dog = difference_of_gaussians(fixed_img, 1, 4)
    # Window Images (Fix spectral boundary)
    wfixed_img = fixed_img_dog * window("hann", fixed_img.shape)
    # Transform image to Fourier space
    fixed_img_fs = np.abs(fft.fftshift(fft.fft2(wfixed_img))) ** 2
    # Compute Log Polar Transform
    radius = fixed_img_fs.shape[0] // 8  # Low Pass
    warped_fixed_img_fs = warp_polar(
        fixed_img_fs,
        radius=radius,
        output_shape=fixed_img_fs.shape,
        scaling="log",
    )
    # Only use half of FFT, because it's symmetrical
    warped_fixed_img_fs = warped_fixed_img_fs[: fixed_img_fs.shape[0] // 2, :]

    # Now prepare for rotating original images,
    #   and searching for translations.
    # We start back at the raw fixed_img.
    twfixed_img = fixed_img * window("hann", fixed_img.shape)

    # Register image `m` against images[0]
    for m in range(1, len(images)):
        # Get the image to register
        regis_img = images[m]

        # Reflect images when necessary
        if reflection[m]:
            regis_img = np.flipud(regis_img)

        # Difference of Gaussians (Band Filter)
        regis_img_dog = difference_of_gaussians(regis_img, 1, 4)

        # Window Images (Fix spectral boundary)
        wregis_img = regis_img_dog * window("hann", regis_img.shape)

        # Transform image to Fourier space
        regis_img_fs = np.abs(fft.fftshift(fft.fft2(wregis_img))) ** 2

        # Compute Log Polar Transform
        warped_regis_img_fs = warp_polar(
            regis_img_fs,
            radius=radius,  # Low Pass
            output_shape=fixed_img_fs.shape,
            scaling="log",
        )

        # Only use half of FFT, because it's symmetrical
        warped_regis_img_fs = warped_regis_img_fs[: fixed_img_fs.shape[0] // 2, :]

        # Compute the Cross_Correlation to estimate rotation
        # Note that _phase_cross_correlation uses the mangnitudes (abs()),
        #  ie it is using both freq and phase information.
        cross_correlation, _ = _phase_cross_correlation(
            warped_fixed_img_fs, warped_regis_img_fs
        )

        # Rotating Cartesian space translates the angular log polar component.
        # Scaling Cartesian space translates the radial log polar component.
        # In common image resgistration problems, both components are used
        #   to simultaneously estimate scaling and rotation.
        # Since we are not currently concerned with scaling transformation,
        #   disregard the second axis of the `cross_correlation` returned by
        #   `_phase_cross_correlation`.
        cross_correlation_score = cross_correlation[:, 0].ravel()

        # Recover the angle from index representing maximal cross_correlation
        recovered_angle_degrees = (360 / regis_img_fs.shape[0]) * np.argmax(
            cross_correlation_score
        )

        # The recovered angle represents an estimate of the rotation from reference to image[m].
        # The registration angle for image[m],
        #   the angle to apply to image[m] to register with reference,
        #   would be the negation of this,
        r = -1 * recovered_angle_degrees

        # For now, try the hack below, attempting two cases ...
        # Some papers mention running entire algos /twice/,
        #   when admitting reflections, so this hack is not
        #   the worst you could do :).
        # Hack
        regis_img_estimated = rotate(regis_img, r)
        regis_img_rotated_p180 = rotate(regis_img, r + 180)
        da = np.dot(fixed_img[mask], regis_img_estimated[mask])
        db = np.dot(fixed_img[mask], regis_img_rotated_p180[mask])
        if db > da:
            regis_img_estimated = regis_img_rotated_p180
            r += 180

        # Assign estimated rotations results
        rotations[m] = r * np.pi / 180  # Convert to radians

        if do_cross_corr_translations:
            # Prepare for searching over translations using cross-correlation with the rotated image.
            twregis_img = regis_img_estimated * window("hann", regis_img.shape)
            cross_correlation, shift = _phase_cross_correlation(
                twfixed_img, twregis_img
            )

            # Compute the shifts as integer number of pixels,
            shift_x, shift_y = int(shift[1]), int(shift[0])
            # then apply the shifts
            regis_img_estimated = np.roll(regis_img_estimated, shift_y, axis=0)
            regis_img_estimated = np.roll(regis_img_estimated, shift_x, axis=1)
            # Assign estimated shift to results
            shifts[m] = shift[::-1].astype(int)

        else:
            shift = None  # For logger line

        # Estimated `corr` metric
        corr = np.dot(fixed_img[mask], regis_img_estimated[mask])
        correlations[m] = corr

    # Cleanup some cached stuff for this class
    __cache.pop(id(warped_fixed_img_fs), None)
    __cache.pop(id(twfixed_img), None)

    return rotations, shifts, correlations



"""
Selecting the "best" classes is an area of active research.

Here we provide an abstract base class with two naive approaches as
concrete implementations.

`RandomClassSelector` will select random indices from across the
entire dataset, with RNG controlled by `seed`.

`TopClassSelector' will select the first `n_classes` in order.  This
may be useful for debugging and development.

Additionally we provide a few methods that have been used
historically, along with a few classes which should aid in
constructing new methods.
"""

import logging
from abc import ABC, abstractmethod
from heapq import heappush, heappushpop
from operator import eq, le

import numpy as np

from aspire.classification import Averager2D
from aspire.image import Image
from aspire.utils import grid_2d

logger = logging.getLogger(__name__)


class ClassSelector(ABC):
    """
    Abstract interface for class selection.
    """

    @abstractmethod
    def _select(self, classes, reflections, distances):
        """
        Using the provided arguments, returns an array representing
        an index into selected `classes`.

        This is the method developers should implement for a custom algorithm.

        :param classes: (n_img, n_nbor) array of image indices
        :param reflections: (n_img, n_nbor) boolean array of reflections between `classes[i][0]` and classes[i][j]`
        :param distances: (n_img, n_nbor) array of distances between `classes[i][0]` and classes[i][j]`

        :return: array of indices into `classes`
        """

    @property
    def quality_scores(self):
        """
        All `ClassSelector` should assign a quality score array the
        same length as the selection output.

        Function range is currently not limited, but [0,1] is
        favorable.  Currently there is not an expectation that one
        quality scoring system relates to another, or that the score
        is a proper metric. Quality scores are only required to be a
        self consistent ordering.

        For subclasses like TopClassSelector and RandomClassSelector
        where no quality information is derived, the associated
        `_quality_scores` should be set to zeros by `_select`.
        """
        # Remind the user to run selection first.  We could run it
        # automatically, but some of the class selectors may take a
        # long time.
        if not hasattr(self, "_quality_scores"):
            raise RuntimeError("Must run `.select(...)` to compute quality_scores")
        return self._quality_scores

    def select(self, classes, reflections, distances):
        """
        Using the provided arguments, calls internal `_select`
        method, checks selection is sane, and returns an array representing
        an ordered index into `classes`.

        :param classes: (n_img, n_nbor) array of image indices
        :param reflections: (n_img, n_nbor) boolean array of reflections between `classes[i][0]` and classes[i][j]`
        :param distances: (n_img, n_nbor) array of distances between `classes[i][0]` and classes[i][j]`

        :return: array of indices into `classes`
        """

        # Compute max class id found in the network for sanity
        # checking. Base class defaults to the `n` known from the
        # initializing `Source`.  This `n` may be reduced through
        # class selection.
        if not hasattr(self, "n"):
            self.n = np.max(classes[:, 0]) + 1  # +1 for zero indexing

        # Call the selection logic
        self.selection = self._select(classes, reflections, distances)

        # n_img should not exceed the largest index in first column of `classes`.
        n_img = np.max(classes[:, 0]) + 1  # +1 for zero indexing
        # Check values in selection are in the bounds.
        self._check_selection(self.selection, n_img)

        return self.selection

    def _check_selection(self, selection, n_img, len_operator=eq):
        """
        Check that class `selection` is sane.

        :param selection: selection indices
        :param n_img: number of images available
        :param len_operator: Operation used for comparison.
            Defaults to equality (`eq`).
        """
        # Check length, +1 for zero indexing
        if not len_operator(len(selection), self.n):
            raise ValueError(
                f"Class selection must be {str(len_operator)} {self.n}, got {len(selection)}"
            )

        # Check indices [0, n_img)
        if (np.max(selection) >= n_img) or (np.min(selection) < 0):
            raise ValueError(
                f"Class selection out of bounds [0, {n_img - 1}]"
                f"with [{np.min(selection)}, {np.max(selection)}]"
            )


class TopClassSelector(ClassSelector):
    def _select(self, classes, reflections, distances):
        """
        Returns classes in `Source` order.

        Mainly useful for debugging.
        """
        # Assign uniform quality.
        self._quality_scores = np.zeros(self.n)
        # Return same indices as initial source.
        return np.arange(self.n)


class RandomClassSelector(ClassSelector):
    def __init__(self, seed=None):
        """
        :param seed: RNG seed, de
        """
        self.seed = seed

    def _select(self, classes, reflections, distances):
        """
        Select random classes from the population.
        """
        # Assign uniform quality.
        self._quality_scores = np.zeros(self.n)

        # Instantiate a random Generator
        rng = np.random.default_rng(self.seed)
        # Generate and return indices for random sample
        # +1 for zero indexing
        return rng.choice(self.n, size=self.n, replace=False)


class NeighborVarianceClassSelector(ClassSelector):
    """
    Selects classes based on variances of `distances`.

    Note that `distances` is the Nearest Neighbors distances,
    and in the case of RIR this is a small rotationally invariant feature
    vector.  For methods based on class average images, see subclasses of `GlobalClassSelector`.
    """

    def _select(self, classes, reflections, distances):
        # Compute per class variance
        dist_var = np.var(distances, axis=1)

        # Compute the ordering, ascending
        sorted_class_inds = np.argsort(dist_var)
        # Store the sorted quality scores (maps to selection output).
        self._quality_scores = dist_var[sorted_class_inds]

        # Return indices
        return sorted_class_inds


class DistanceClassSelector(ClassSelector):
    """
    Selects top classes based on lowest mean distance
    as estimated by `distances`.

    Note that `distances` is the Nearest Neighbors distances,
    and in the case of RIR this is a small rotationally invariant feature
    vector.  For methods based on class average images, see subclasses of `GlobalClassSelector`.
    """

    def _select(self, classes, reflections, distances):
        # Compute per class distance.
        # Skips self distance (0).
        dist_mean = np.mean(distances[:, 1:], axis=1)

        # Compute the ordering, ascending
        sorted_class_inds = np.argsort(dist_mean)
        self._quality_scores = dist_mean[sorted_class_inds]

        # Return indices
        return sorted_class_inds


class _HeapItem:
    """
    Organize our heap entries.
    """

    def __init__(self, value, index, image):
        """
        Initialize a heap item.

        :param value: Float value used for heap ranking
        :param index: Image index
        :param image: Image object
        """
        self.value = float(value)
        self.index = int(index)
        self.image = image

    def __lt__(self, other):
        """
        This comparator is used by `heapq` to order and maintain the heap.
        """
        return self.value < other.value

    @staticmethod
    def nbytes(img_size, dtype):
        """
        Computes a rough size of _HeapItem based on assigned
        attributes. Note doesn't include python object overhead.

        :param img_size: Image size in pixels, (img_size, img_size).
        :param dtype: Image datatype.

        :return: Sum of attribute sizes.
        """
        # Allow passing dtype directly, or instance of dtype.
        dtype = np.dtype(dtype)
        return img_size**2 * dtype.itemsize + 16


class GlobalClassSelector(ClassSelector):
    """
    Extends ClassSelector for methods that require
    passing over all class average images.
    """

    def __init__(self, averager, quality_function, heap_size_limit_bytes=2e9):
        """
        Initializes a GlobalClassSelector.

        Because GlobalClassSelectors must compute all class averages,
        a heap cache maintains the top class averages as scored by
        `quality_function`.  If you have the memory, recommend setting
        the cache to be > n_classes*img_size*img_size*img.dtype.

        :param averager: An Averager2D subclass.
        :param quality_function: Function that takes an image and
            returns numeric quality score.  This score will be used to
            sort the classes.  User's may provide a callable function,
            but extending `ImageQualityFunction` is recommended.  For
            example, this module provides methods for variance and SNR
            based quality.
        :param heap_size_limit_bytes: Max heap size in Bytes.
            Defaults 2GB, 0 will disable.
        """
        self.averager = averager
        if not isinstance(self.averager, Averager2D):
            raise ValueError(
                f"`averager` should be instance of `Averager2D`, found {self.averager}."
            )

        self._quality_function = quality_function
        if not callable(self._quality_function):
            raise ValueError(
                "`quality_function` must be a callable function.  See ImageQualityFunction."
            )

        self.n = self.averager.src.n
        # We should eventually compute all `n` entries,
        # but start with identifying missing values as nan.
        self._quality_scores = np.full(self.n, fill_value=float("nan"))

        # To be used for heapq (score, img_id, avg_im)
        # Note the default implementation is min heap, so `pop` returns smallest value.
        self.heap = []

        self._heap_limit_bytes = int(heap_size_limit_bytes)
        self._heap_item_size = self.averager.src.L**2 * self.averager.dtype.itemsize

    @property
    def _heap_size(self):
        """
        Return estimate of heap_size.  Note this is not the exact
        heap_size, as it doesn't include the overhead for python
        objects.
        """
        n = len(self.heap)
        item_size = _HeapItem.nbytes(
            img_size=self.averager.composite_basis.nres, dtype=self.averager.dtype
        )

        return n * item_size

    @property
    def heap_ids(self):
        """
        Return the image ids currently in the heap.
        """
        return [item.index for item in self.heap]

    @property
    def heap_idx_map(self):
        """
        Return map of image ids to heap position currently in the heap.
        """
        return {item.index: i for i, item in enumerate(self.heap)}

    def _select(self, classes, reflections, distances):
        for i, im in enumerate(self.averager.average(classes, reflections)):
            quality_score = self._quality_function(im)

            # Assign in global quality score array
            self._quality_scores[i] = quality_score

            # Skip heap if single item is larger than heap limit.
            # Implies `self._heap_limit_bytes = 0` disables heap.
            if self._heap_item_size > self._heap_limit_bytes:
                continue

            # Create cachable payload item
            item = _HeapItem(quality_score, i, im)

            if self._heap_item_size + self._heap_size < self._heap_limit_bytes:
                # There is space to push another entry onto heap
                heappush(self.heap, item)
            elif self._heap_size < self._heap_limit_bytes:
                # Heap is currently out of space,
                # pop and throw away the worst entry
                # after updating (pushing to) the heap.
                _ = heappushpop(self.heap, item)

        # Now that we have computed the global quality_scores,
        # the selection ordering can be applied, descending
        sorted_class_inds = np.argsort(self._quality_scores)[::-1]
        self._quality_scores = self._quality_scores[sorted_class_inds]
        return sorted_class_inds


# TODO: When a consistent measure of distance is implemented by
# preceeding components we can implement exclusion based on neighbor
# distances or other ideas (VDM?) as different ClassRepulsion like
# classes.
class GreedyClassRepulsionMixin:
    """
    Mixin to overload class selection based on excluding
    classes we've already seen as neighbors of another class.

    If the classes are well sorted (by some measure of quality),
    we assume the best representation is the first seen.
    """

    def __init__(self, *args, **kwargs):
        """
        Sets optional `exclude_k`. All other args and **kwargs are
        passed to super().

        GreedyClassRepulsionMixin is similar to `cryo_select_subset` from
        MATLAB, but MATLAB found `exclude_k` iteratively based on a
        desired result set size.

        :param exclude_k: Number of neighbors from each class to
            exclude.  Defaults to all neighbors.
        """
        # Pop of the parameter unique to GreedyClassRepulsionMixin.
        self.exclude_k = kwargs.pop("exclude_k", None)

        # Instantiate an empty set to hold our excluded indices.
        self.excluded = set()

        # Pass everything else through to the super __init__.
        super().__init__(*args, **kwargs)

    def _select(self, classes, reflections, distances):
        """
        Overload the ClassSelector._select
        """
        # Get the indices sorted by the super's `_select` method.
        sorted_inds = super()._select(classes, reflections, distances)

        # If exclude_k is not provided, default to exluding all
        # neighbors seen.
        k = self.exclude_k or classes.shape[-1]

        results = []
        for i in sorted_inds:
            # Skip when this class's base image has been seen in a
            # prior class.
            if i in self.excluded:
                continue

            # Get the images in this class `i`, and add class images
            # up to `k` the exclusion list.
            self.excluded.update(classes[i, :k])

            results.append(i)

        return np.array(results, dtype=int)

    def _check_selection(self, selection, n_img):
        """
        Check that class `selection` is sane.

        Repulsion can reduce the number of classes (dramatically),
        so we need to adjust the checking operation.

        :param selection: selection indices
        :param n_img: number of images available
        """
        return super()._check_selection(selection, n_img, len_operator=le)


class NeighborVarianceWithRepulsionClassSelector(
    GreedyClassRepulsionMixin, NeighborVarianceClassSelector
):
    """
    Selects top classes based on highest contrast with GreedyClassRepulsionMixin.
    """


class GlobalWithRepulsionClassSelector(GreedyClassRepulsionMixin, GlobalClassSelector):
    """
    Extends ClassSelector for methods that require
    passing over all class average images and also GreedyClassRepulsionMixin.
    """


class ImageQualityFunction(ABC):
    """
    A callable image quality scoring function.

    The main advantage to using this class is to gain access to a grid
    caching and Image/Numpy conversion.
    """

    _grid_cache = {}

    @abstractmethod
    def _function(self, img):
        """
        User defined scoring function.  Function domain and range is
        currently not limited, but [0,1] is favorable range.

        Developers can use the self._grid_cache for access to a
        grid_2d instance matching resolution of img.

        :param img: 2d Numpy array :returns: Image quality score
        """

    def __call__(self, img):
        """
        Given an image instance or a 2D np array,
        calls the instance's weight function.

        :param img: `Image`, 2d Numpy array,
            or 3d array where slow axis is stack axis.
            When called with an image stack>1, an array is returned.
        :returns: Image quality score(s)
        """

        if isinstance(img, Image):
            img = img.asnumpy()

        # Allow the function to be used on a single image presented as
        # a 2d Numpy array, as well as a stack.
        if img.ndim == 2:
            stack_len = 0  # singleton
            img = img[np.newaxis, :, :]
        else:
            stack_len = img.shape[0]

        # Image size.
        L = img.shape[-1]

        # Generate grid on first call as needed.
        self._grid_cache.setdefault(L, grid_2d(L, dtype=img.dtype))

        # Call the function over img stack.
        res = np.fromiter(map(self._function, img), dtype=img.dtype)

        # Return singleton when given a singleton (2d array).
        if stack_len == 0:
            res = res[0]

        return res


class VarianceImageQualityFunction(ImageQualityFunction):
    """
    Computes the variance of pixels.
    """

    def _function(self, img):
        """
        Scoring function based on variance.

        :param img: Input image as 2d Numpy array.

        :return: Pixel variance.
        """
        return np.var(img)


class BandedSNRImageQualityFunction(ImageQualityFunction):
    """
    Computes the ratio of variance of central pixels of image to
    pixels in a configurable outer band.
    """

    def _function(self, img, center_radius=0.5, outer_band_start=0.8, outer_band_end=1):
        """Configurable scoring function.

        :param img: Input image as 2d Numpy array.
        :param center_radius: Proportion of image radius defining the center.
        :param outer_band_start: Proportion of image radius defining
            the inner boundary of band.
        :param outer_band_end: Proportion of image radius defining the
            outer boundary of band.  Must be larger than outer_band_start.

        :return: Ratio central variance to outer band variance.
        """
        # Image Size
        L = img.shape[-1]

        # Look up the grid.
        grid = self._grid_cache[L]

        center_mask = grid["r"] < center_radius
        outer_mask = (grid["r"] > outer_band_start) & (grid["r"] < outer_band_end)
        if not np.any(outer_mask):
            # outer_mask is empty, need to correct outer_band_{start,end}
            raise RuntimeError(
                f"Band of ({outer_band_start}, {outer_band_end}) empty for image size {L}, adjust band boundaries."
            )

        return np.var(img[center_mask]) / np.var(img[outer_mask])


class BandpassImageQualityFunction(ImageQualityFunction):
    """
    Replicate behavior of MATLAB `cryo_sort_stack_bandpass` method.
    """

    # TODO, sort by polar Fourier bandpass of Image.
    # This was implemented in MATLAB, but appears unused.
    # Note, we may derive a similar method for non global use using bispectrum
    #   (if we do bookkeeping during compression).


class WeightedImageQualityMixin(ABC):
    """
    Extends ImageQualityFunction with a radial grid weighted function
    for use in user defined `_function` calls.
    """

    def __init__(self):
        # Each weight function will need a seperate cache.
        self._weights_cache = {}
        super().__init__()

    @abstractmethod
    def _weight_function(self, r):
        """
        Take a 1d radial weight function.
        The function is expected to be defined [0,sqrt(2)].
        Function range is not limited, but should probably be [0,1].

        :param r: Radius (grid).
        :return: 2D grid of weights
        """

    def weights(self, L):
        """
        Returns  2D array of weights for a given resolution L.
        Computes and caches on first request.

        :param L: resolution pixels

        :return: 2d weight array for LxL grid.
        """
        grid = self._grid_cache[L]
        # frompyfunc performs the broadcast.
        return self._weights_cache.setdefault(
            L, np.frompyfunc(self._weight_function, 1, 1)(grid["r"])
        )

    def _function(self, img):
        """
        Apply weights to `img` before calling underlying `_function`.

        Users may override this to use weights in more sophisticated
        manner.
        """
        img = img * self.weights(img.shape[-1])
        return super()._function(img)


# These classes are provided as helpers/examples.
class RampWeightedImageQualityMixin(WeightedImageQualityMixin):
    """
    ImageQualityMixin to apply a linear ramp.
    """

    def _weight_function(self, r):
        # linear ramp
        return np.max(1 - r, 0)


class BumpWeightedImageQualityMixin(WeightedImageQualityMixin):
    """
    ImageQualityMixin to apply a [0,1] bump function.
    """

    def _weight_function(self, r):
        # bump function (normalized to [0,1]).
        if r >= 1:
            return 0
        else:
            return np.exp(-1 / (1 - r**2) + 1)


class BumpWeightedVarianceImageQualityFunction(
    BumpWeightedImageQualityMixin, VarianceImageQualityFunction
):
    """
    Computes the variance of pixels after weighting with Bump function.
    """


class RampWeightedVarianceImageQualityFunction(
    RampWeightedImageQualityMixin, VarianceImageQualityFunction
):
    """
    Computes the variance of pixels after weighting with Ramp function.
    """



import logging
from functools import partial

import numpy as np
import scipy.sparse.linalg
from scipy.linalg import norm
from scipy.sparse.linalg import LinearOperator

from aspire.nufft import anufft
from aspire.numeric import fft
from aspire.operators import evaluate_src_filters_on_grid
from aspire.reconstruction import Estimator, FourierKernel, MeanEstimator
from aspire.utils import (
    make_symmat,
    symmat_to_vec_iso,
    trange,
    vec_to_symmat_iso,
    vecmat_to_volmat,
    volmat_to_vecmat,
)
from aspire.volume import Volume, rotated_grids

logger = logging.getLogger(__name__)


class CovarianceEstimator(Estimator):
    def __init__(self, *args, **kwargs):
        if "mean_kernel" in kwargs:
            self.mean_kernel = kwargs.pop("mean_kernel")
        super().__init__(*args, **kwargs)

    def __getattr__(self, name):
        """Lazy attributes instantiated on first-access"""

        if name == "mean_kernel":
            mean_kernel = self.mean_kernel = MeanEstimator(self.src, self.basis).kernel
            return mean_kernel
        return super(CovarianceEstimator, self).__getattr__(name)

    def compute_kernel(self):
        # TODO: Most of this stuff is duplicated in MeanEstimator - move up the hierarchy?
        n = self.src.n
        L = self.src.L
        _2L = 2 * self.src.L

        kernel = np.zeros((_2L, _2L, _2L, _2L, _2L, _2L), dtype=self.dtype)
        sq_filters_f = np.square(evaluate_src_filters_on_grid(self.src))

        for i in trange(0, n, self.batch_size):
            _range = np.arange(i, min(n, i + self.batch_size))
            pts_rot = rotated_grids(L, self.src.rotations[_range, :, :])
            weights = sq_filters_f[:, :, _range]
            weights *= self.src.amplitudes[_range] ** 2

            if L % 2 == 0:
                weights[0, :, :] = 0
                weights[:, 0, :] = 0

            # TODO: This is where this differs from MeanEstimator
            pts_rot = np.moveaxis(pts_rot[::-1], 1, 0).reshape(-1, 3, L**2)
            weights = weights.T.reshape((-1, L**2))

            batch_n = weights.shape[0]
            factors = np.zeros((batch_n, _2L, _2L, _2L), dtype=self.dtype)

            for j in range(batch_n):
                factors[j] = anufft(weights[j], pts_rot[j], (_2L, _2L, _2L), real=True)

            factors = Volume(factors).to_vec()
            kernel += vecmat_to_volmat(factors.T @ factors) / (n * L**8)

        # Ensure symmetric kernel
        kernel[0, :, :, :, :, :] = 0
        kernel[:, 0, :, :, :, :] = 0
        kernel[:, :, 0, :, :, :] = 0
        kernel[:, :, :, 0, :, :] = 0
        kernel[:, :, :, :, 0, :] = 0
        kernel[:, :, :, :, :, 0] = 0

        logger.info("Computing non-centered Fourier Transform")
        kernel = fft.mdim_ifftshift(kernel, range(0, 6))
        kernel_f = fft.fftn(kernel)
        # Kernel is always symmetric in spatial domain and therefore real in Fourier
        kernel_f = np.real(kernel_f)

        return FourierKernel(kernel_f, centered=False)

    def estimate(self, mean_vol, noise_variance, tol=1e-5, regularizer=0):
        logger.info("Running Covariance Estimator")
        b_coeff = self.src_backward(mean_vol, noise_variance)
        est_coeff = self.conj_grad(b_coeff, tol=tol, regularizer=regularizer)
        covar_est = self.basis.mat_evaluate(est_coeff)
        covar_est = vecmat_to_volmat(make_symmat(volmat_to_vecmat(covar_est)))
        return covar_est

    def conj_grad(self, b_coeff, tol=1e-5, regularizer=0):
        b_coeff = symmat_to_vec_iso(b_coeff)
        N = b_coeff.shape[0]
        kernel = self.kernel

        if regularizer > 0:
            kernel += regularizer

        operator = LinearOperator(
            (N, N),
            matvec=partial(self.apply_kernel, kernel=kernel, packed=True),
            dtype=self.dtype,
        )
        if self.precond_kernel is None:
            M = None
        else:
            precond_kernel = self.precond_kernel
            if regularizer > 0:
                precond_kernel += regularizer
            M = LinearOperator(
                (N, N),
                matvec=partial(self.apply_kernel, kernel=precond_kernel, packed=True),
                dtype=self.dtype,
            )

        target_residual = tol * norm(b_coeff)

        def cb(xk):
            logger.info(
                f"Delta {norm(b_coeff - self.apply_kernel(xk, packed=True))} (target {target_residual})"
            )

        x, info = scipy.sparse.linalg.cg(
            operator, b_coeff, M=M, callback=cb, tol=tol, atol=0
        )

        if info != 0:
            raise RuntimeError("Unable to converge!")
        return vec_to_symmat_iso(x)

    def apply_kernel(self, coeff, kernel=None, packed=False):
        """
        Applies the kernel represented by convolution

        :param coeff: The volume matrix (6 dimensions) to be convolved (but see the `packed` argument below).
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :param packed: whether the `coeff` matrix represents an isometrically mapped packed vector,
            through the `symmat_to_vec_iso` function. In this case, the function expands `coeff` into a symmetric
            matrix internally, and returns a packed vector in return.
        :return: The result of evaluating `coeff` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis. If `packed` is True, then the isometrically mapped packed
            vector is returned instead.
        """
        if kernel is None:
            kernel = self.kernel
        if packed:
            coeff = vec_to_symmat_iso(coeff)

        result = self.basis.mat_evaluate_t(
            kernel.convolve_volume_matrix(self.basis.mat_evaluate(coeff))
        )
        return symmat_to_vec_iso(result) if packed else result

    def src_backward(self, mean_vol, noise_variance, shrink_method=None):
        """
        Apply adjoint mapping to source

        :return: The sum of the outer products of the mean-subtracted images in `src`, corrected by the expected noise
            contribution and expressed as coefficients of `basis`.
        """
        covar_b = np.zeros(
            (self.src.L, self.src.L, self.src.L, self.src.L, self.src.L, self.src.L),
            dtype=self.dtype,
        )

        for i in range(0, self.src.n, self.batch_size):
            im = self.src.images[i : i + self.batch_size]
            batch_n = im.n_images
            im_centered = im - self.src.vol_forward(mean_vol, i, self.batch_size)

            im_centered_b = np.zeros(
                (batch_n, self.src.L, self.src.L, self.src.L), dtype=self.dtype
            )
            for j in range(batch_n):
                im_centered_b[j] = self.src.im_backward(im_centered[j], i + j)
            im_centered_b = Volume(im_centered_b).to_vec()

            covar_b += vecmat_to_volmat(im_centered_b.T @ im_centered_b) / self.src.n

        covar_b_coeff = self.basis.mat_evaluate_t(covar_b)
        return self._shrink(covar_b_coeff, noise_variance, shrink_method)

    def _shrink(self, covar_b_coeff, noise_variance, method=None):
        """
        Shrink covariance matrix

        :param covar_b_coeff: Outer products of the mean-subtracted images
        :param noise_variance: Noise variance
        :param method: One of None/'frobenius_norm'/'operator_norm'/'soft_threshold'
        :return: Shrunk covariance matrix
        """
        assert method in (
            None,
            "frobenius_norm",
            "operator_norm",
            "soft_threshold",
        ), "Unsupported shrink method"

        An = self.basis.mat_evaluate_t(self.mean_kernel.toeplitz())
        if method is None:
            covar_b_coeff -= noise_variance * An
        else:
            raise NotImplementedError("Only default shrink method supported.")

        return covar_b_coeff



import numpy as np


class MrcStats:
    def __init__(self):
        """
        Instantiate an empty instance ready to receive slices of arrays.

        :return: MrcStats instance
        """

        self.amin = np.inf
        self.amax = -np.inf
        self.asum = 0.0
        self.asize = 0
        self.asum2 = 0.0

    def push(self, array_slice):
        """
        Incrementally add contribution of array_slice to stats.

        :param array_slice: An ndarray to contribute to stats.
        """

        self.amin = min(self.amin, np.min(array_slice))
        self.amax = max(self.amax, np.max(array_slice))
        # For mean we'll do div when called.
        self.asum += np.sum(array_slice)
        self.asum2 += np.sum(np.square(array_slice))
        self.asize += np.size(array_slice)

    @property
    def amean(self):
        """
        Compute the curren mean.
        """

        return self.asum / self.asize

    @property
    def arms(self):
        """
        Compute the current standard deviation.

        We use the simplest method.
        """

        var = self.asum2 / self.asize - (self.asum / self.asize) ** 2
        return np.sqrt(var)

    def update_header(self, mrcobj):
        """
        Assign the statistics to `mrcobj` header.

        :param mrcobj: The `mrcfile` instance we want stats written to.
        """

        mrcobj.header.dmin = self.amin.astype(np.float32)
        mrcobj.header.dmax = self.amax.astype(np.float32)
        mrcobj.header.dmean = self.amean.astype(np.float32)
        mrcobj.header.rms = self.arms.astype(np.float32)
